#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
from datetime import datetime, timedelta
from typing import List, Literal, Optional, Union, get_args

import humps
from hopsworks_common import (
    alert,
    alert_receiver,
    client,
    decorators,
    triggered_alert,
    usage,
)
from hopsworks_common.engine import alerts_engine


_GLOBAL_RECEIVER_PREFIX = "global-receiver"
_API_MATCH = {"key": "api", "value": "alerts-api"}

_PROJECT_JOB_STATUS_ARG = Literal[
    "job_finished",
    "job_failed",
    "job_killed",
    "job_long_running",
]

_PROJECT_JOB_STATUS = get_args(_PROJECT_JOB_STATUS_ARG)

_PROJECT_FS_STATUS_ARG = Literal[
    "validation_success",
    "validation_warning",
    "validation_failure",
    "feature_validation_success",
    "feature_validation_warning",
    "feature_validation_failure",
    "feature_monitor_shift_undetected",
    "feature_monitor_shift_detected",
]
_PROJECT_FS_STATUS = get_args(_PROJECT_FS_STATUS_ARG)

_SEVERITY_ARG = Literal["warning", "critical", "info"]
_SEVERITY = get_args(_SEVERITY_ARG)

_SERVICES_ARG = Literal["Featurestore", "Jobs"]
_SERVICES = get_args(_SERVICES_ARG)

_JOB_STATUS_ARG = Literal[
    "finished",
    "failed",
    "killed",
    "long_running",
    "job_finished",
    "job_failed",
    "job_killed",
    "job_long_running",
]
_JOB_STATUS = get_args(_JOB_STATUS_ARG)

_VALIDATION_STATUS_ARG = Literal[
    "success",
    "warning",
    "failure",
    "validation_success",
    "validation_warning",
    "validation_failure",
    "feature_validation_success",
    "feature_validation_warning",
    "feature_validation_failure",
]
_VALIDATION_STATUS = get_args(_VALIDATION_STATUS_ARG)

_MONITORING_STATUS_ARG = Literal[
    "feature_monitor_shift_undetected",
    "feature_monitor_shift_detected",
]
_MONITORING_STATUS = get_args(_MONITORING_STATUS_ARG)


class AlertsApi:
    def __init__(self):
        self._log = logging.getLogger(__name__)

    @usage.method_logger
    def get_alerts(self) -> List[alert.ProjectAlert]:
        """
        Get all project alerts.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        alert = alerts_api.get_alerts(alert_id=1)

        ```
        # Returns
            `List[ProjectAlert]`: List of ProjectAlert objects.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "service", "alerts"]
        headers = {"content-type": "application/json"}
        return alert.ProjectAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    @decorators.catch_not_found(
        "hopsworks_common.alert.ProjectAlert", fallback_return=None
    )
    def get_alert(self, alert_id: int) -> Optional[alert.ProjectAlert]:
        """
        Get a specific project alert by ID.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        alert = alerts_api.get_alert(alert_id=1)

        ```
        # Arguments
            alert_id: The ID of the alert to retrieve.
        # Returns
            `ProjectAlert`: The ProjectAlert object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "service", "alerts", alert_id]
        headers = {"content-type": "application/json"}
        return alert.ProjectAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    def get_job_alerts(self, job_name: str) -> List[alert.JobAlert]:
        """
        Get all job alerts.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        job_alerts = alerts_api.get_job_alerts(job_name="my_job")

        ```
        # Arguments
            job_name: The name of the job.
        # Returns
            `List[JobAlert]`: List of JobAlert objects.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "jobs", job_name, "alerts"]
        headers = {"content-type": "application/json"}
        return alert.JobAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    @decorators.catch_not_found("hopsworks_common.alert.JobAlert", fallback_return=None)
    def get_job_alert(self, job_name: str, alert_id: int) -> Optional[alert.JobAlert]:
        """
        Get a specific job alert by ID.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        job_alerts = alerts_api.get_job_alert(job_name="my_job", alert_id=1)

        ```
        # Arguments
            job_name: The name of the job.
            alert_id: The ID of the alert to retrieve.
        # Returns
            `JobAlert`: The JobAlert object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "jobs",
            job_name,
            "alerts",
            alert_id,
        ]
        headers = {"content-type": "application/json"}
        return alert.JobAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    def get_feature_group_alerts(
        self,
        feature_store_id: int,
        feature_group_id: int,
    ) -> List[alert.FeatureGroupAlert]:
        """
        Get all feature group alerts.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        feature_group_alerts = alerts_api.get_feature_group_alerts(feature_store_id=1, feature_group_id=1)

        ```
        # Arguments
            feature_store_id: The ID of the feature store.
            feature_group_id: The ID of the feature group.
        # Returns
            `List[FeatureGroupAlert]`: List of FeatureGroupAlert objects.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featuregroups",
            feature_group_id,
            "alerts",
        ]
        headers = {"content-type": "application/json"}
        return alert.FeatureGroupAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    @decorators.catch_not_found(
        "hopsworks_common.alert.FeatureGroupAlert", fallback_return=None
    )
    def get_feature_group_alert(
        self,
        feature_store_id: int,
        feature_group_id: int,
        alert_id: int,
    ) -> Optional[alert.FeatureGroupAlert]:
        """
        Get a specific feature group alert by ID.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        feature_group_alerts = alerts_api.get_feature_group_alert(feature_store_id=1, feature_group_id=1, alert_id=1)

        ```
        # Arguments
            feature_store_id: The ID of the feature store.
            feature_group_id: The ID of the feature group.
            alert_id: The ID of the alert to retrieve.
        # Returns
            `FeatureGroupAlert`: The FeatureGroupAlert object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featuregroups",
            feature_group_id,
            "alerts",
            alert_id,
        ]
        headers = {"content-type": "application/json"}
        return alert.FeatureGroupAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    def get_feature_view_alerts(
        self,
        feature_store_id: int,
        feature_view_name: str,
        feature_view_version: int,
    ) -> List[alert.FeatureViewAlert]:
        """
        Get all feature view alerts.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        feature_view_alerts = alerts_api.get_feature_view_alerts(feature_store_id=1, feature_view_name="my_feature_view", feature_view_version=1, alert_id=1)

        ```
        # Arguments
            feature_store_id: The ID of the feature store.
            feature_view_name: The name of the feature view.
            feature_view_version: The version of the feature view.
        # Returns
            `List[FeatureViewAlert]`: List of FeatureViewAlert objects.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featureview",
            feature_view_name,
            "version",
            feature_view_version,
            "alerts",
        ]
        headers = {"content-type": "application/json"}
        return alert.FeatureViewAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    @decorators.catch_not_found(
        "hopsworks_common.alert.FeatureViewAlert", fallback_return=None
    )
    def get_feature_view_alert(
        self,
        feature_store_id: int,
        feature_view_name: str,
        feature_view_version: int,
        alert_id: int,
    ) -> Optional[alert.FeatureViewAlert]:
        """
        Get a specific feature view alert by ID.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        feature_view_alerts = alerts_api.get_feature_view_alert(feature_store_id=1, feature_view_name="my_feature_view", feature_view_version=1, alert_id=1)

        ```
        # Arguments
            feature_store_id: The ID of the feature store.
            feature_view_name: The name of the feature view.
            feature_view_version: The version of the feature view.
            alert_id: The ID of the alert to retrieve.
        # Returns
            `FeatureViewAlert`: The FeatureViewAlert object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featureview",
            feature_view_name,
            "version",
            feature_view_version,
            "alerts",
            alert_id,
        ]
        headers = {"content-type": "application/json"}
        return alert.FeatureViewAlert.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    def create_project_alert(
        self,
        receiver: str,
        status: Union[_PROJECT_FS_STATUS_ARG, _PROJECT_JOB_STATUS_ARG],
        severity: _SEVERITY_ARG,
        service: _SERVICES_ARG,
        threshold=0,
    ) -> alert.ProjectAlert:
        """
        Create a new alert.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        new_alert = alerts_api.create_project_alert(receiver="email", status="job_finished", severity="warning", service="Jobs")

        ```
        # Arguments
            status: The status that will trigger the alert (job_finished, job_failed, job_killed, job_long_running, feature_validation_success, feature_validation_warning, feature_validation_failure, feature_monitor_shift_undetected, feature_monitor_shift_detected).
            severity: The severity of the alert (warning, critical, info).
            receiver: The receiver of the alert (e.g., email, webhook).
            service: The service associated with the alert (Featurestore, Jobs).
            threshold: The threshold for the alert (default is 0).
        # Returns
            `ProjectAlert`: The created ProjectAlert object.
        # Raises
            `ValueError`: If the service is not Featurestore or Jobs, or if the status is not valid for the specified service.
            `ValueError`: If the severity is not valid.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if service not in _SERVICES:
            raise ValueError(f"Service must be one of the following {_SERVICES}.")
        if status not in _PROJECT_FS_STATUS + _PROJECT_JOB_STATUS:
            raise ValueError(
                f"Status must be one of the following: {_PROJECT_FS_STATUS + _PROJECT_JOB_STATUS}."
            )
        if severity not in _SEVERITY:
            raise ValueError(f"Severity must be one of the following: {_SEVERITY}.")
        if service == "Featurestore" and status in _PROJECT_JOB_STATUS:
            raise ValueError(
                f"Featurestore service does not support job alerts. Supported values are {_PROJECT_FS_STATUS}."
            )
        if service == "Jobs" and status in _PROJECT_FS_STATUS:
            raise ValueError(
                f"Jobs service does not support featurestore alerts. Supported values are {_PROJECT_JOB_STATUS}."
            )

        # feature_validation_ prefix is added for readablity in the API
        if status.startswith("feature_validation_"):
            status = status.replace("feature_validation_", "validation_")

        _client = client.get_instance()
        receiver = self._fix_receiver_name(receiver, _client._project_name)
        path_params = ["project", _client._project_id, "service", "alerts"]
        alert_data = {
            "status": status.upper(),
            "severity": severity.upper(),
            "receiver": receiver,
            "service": service.upper(),
            "threshold": threshold,
        }
        headers = {"content-type": "application/json"}
        return alert.ProjectAlert.from_response_json(
            _client._send_request(
                "POST", path_params, data=json.dumps(alert_data), headers=headers
            )
        )

    @usage.method_logger
    def create_feature_group_alert(
        self,
        feature_store_id: int,
        feature_group_id: int,
        receiver: str,
        status: Union[_VALIDATION_STATUS_ARG, _MONITORING_STATUS_ARG],
        severity: _SEVERITY_ARG,
    ) -> alert.FeatureGroupAlert:
        """
        Create a new feature group alert.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        new_alert = alerts_api.create_feature_group_alert(67, 1, receiver="email", status="feature_validation_warning", severity="warning")

        ```
        # Arguments
            feature_store_id: The ID of the feature store.
            feature_group_id: The ID of the feature group.
            receiver: The receiver of the alert (e.g., email, webhook).
            status: The status that will trigger the alert (feature_validation_success, feature_validation_warning, feature_validation_failure, feature_monitor_shift_undetected, feature_monitor_shift_detected).
            severity: The severity of the alert (warning, critical, info).
        # Returns
            `FeatureGroupAlert`: The created FeatureGroupAlert object.
        # Raises
            `ValueError`: If the status is not valid.
            `ValueError`: If the severity is not valid.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        if status not in _VALIDATION_STATUS + _MONITORING_STATUS:
            raise ValueError(
                f"Status must be one of the following: {_VALIDATION_STATUS + _MONITORING_STATUS}."
            )

        # feature_validation_ prefix is added for readablity in the API
        if status.startswith("feature_validation_"):
            status = status.replace("feature_validation_", "")

        # validation_ prefix is added to match the project created alert API
        if status.startswith("validation_"):
            status = status.replace("validation_", "")

        if severity not in _SEVERITY:
            raise ValueError(f"Severity must be one of the following: {_SEVERITY}.")
        _client = client.get_instance()
        receiver = self._fix_receiver_name(receiver, _client._project_name)
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featuregroups",
            feature_group_id,
            "alerts",
        ]

        alert_data = {
            "status": status.upper(),
            "severity": severity.upper(),
            "receiver": receiver,
        }
        headers = {"content-type": "application/json"}
        return alert.FeatureGroupAlert.from_response_json(
            _client._send_request(
                "POST", path_params, data=json.dumps(alert_data), headers=headers
            )
        )

    @usage.method_logger
    def create_feature_view_alert(
        self,
        feature_store_id: int,
        feature_view_name: str,
        feature_view_version: int,
        receiver: str,
        status: _MONITORING_STATUS_ARG,
        severity: _SEVERITY_ARG,
    ) -> alert.FeatureViewAlert:
        """
        Create a new feature view alert.
        ```python
        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        new_alert = alerts_api.create_feature_view_alert(67, "fv", 1, receiver="email", status="feature_monitor_shift_undetected", severity="warning")

        ```
        # Arguments
            feature_store_id: The ID of the feature store.
            feature_view_name: The name of the feature view.
            feature_view_version: The version of the feature view.
            receiver: The receiver of the alert (e.g., email, webhook).
            status: The status that will trigger the alert (feature_monitor_shift_undetected, feature_monitor_shift_detected).
            severity: The severity of the alert (warning, critical, info).
        # Returns
            `FeatureViewAlert`: The created FeatureViewAlert object.
        # Raises
            `ValueError`: if status is not valid.
            `ValueError`: if severity is not valid.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if status not in _MONITORING_STATUS:
            raise ValueError(
                f"Status must be one of the following: {_MONITORING_STATUS}."
            )
        if severity not in _SEVERITY:
            raise ValueError(f"Severity must be one of the following: {_SEVERITY}.")
        _client = client.get_instance()
        receiver = self._fix_receiver_name(receiver, _client._project_name)
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featureview",
            feature_view_name,
            "version",
            feature_view_version,
            "alerts",
        ]
        alert_data = {
            "status": status.upper(),
            "severity": severity.upper(),
            "receiver": receiver,
        }
        headers = {"content-type": "application/json"}
        return alert.FeatureViewAlert.from_response_json(
            _client._send_request(
                "POST", path_params, data=json.dumps(alert_data), headers=headers
            )
        )

    @usage.method_logger
    def create_job_alert(
        self,
        job_name: str,
        receiver: str,
        status: _JOB_STATUS_ARG,
        severity: _SEVERITY_ARG,
    ) -> alert.JobAlert:
        """
        Create a new job alert.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        new_alert = alerts_api.create_job_alert(job_name="my_job", receiver="email", status="finished", severity="warning")

        ```
        # Arguments
            job_name: The name of the job.
            receiver: The receiver of the alert (e.g., email, webhook).
            status: The status of the alert (finished, failed, killed, long_running).
            severity: The severity of the alert (warning, critical, info).
        # Returns
            `JobAlert`: The created JobAlert object.
        # Raises
            `ValueError`: If the job name is not provided.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if status not in _JOB_STATUS:
            raise ValueError(f"Status must be one of the following: {_JOB_STATUS}.")
        if severity not in _SEVERITY:
            raise ValueError(f"Severity must be one of the following: {_SEVERITY}.")

        # job_ prefix is added to match the project created alert API
        if status.startswith("job_"):
            status = status.replace("job_", "")

        _client = client.get_instance()
        receiver = self._fix_receiver_name(receiver, _client._project_name)
        path_params = [
            "project",
            _client._project_id,
            "jobs",
            job_name,
            "alerts",
        ]
        alert_data = {
            "status": status.upper(),
            "severity": severity.upper(),
            "receiver": receiver,
        }
        headers = {"content-type": "application/json"}
        return alert.JobAlert.from_response_json(
            _client._send_request(
                "POST", path_params, data=json.dumps(alert_data), headers=headers
            )
        )

    @usage.method_logger
    def get_alert_receivers(self) -> List[alert_receiver.AlertReceiver]:
        """
        Get all alert receivers.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        alert_receivers = alerts_api.get_alert_receivers()

        ```
        # Returns
            `List[AlertReceiver]`: List of alert receivers.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "alerts", "receivers"]
        query_params = {"expand": True, "global": True}
        headers = {"content-type": "application/json"}
        return alert_receiver.AlertReceiver.from_response_json(
            _client._send_request(
                "GET", path_params, query_params=query_params, headers=headers
            )
        )

    @usage.method_logger
    @decorators.catch_not_found(
        "hopsworks_common.alert_receiver.AlertReceiver", fallback_return=None
    )
    def get_alert_receiver(self, name: str) -> Optional[alert_receiver.AlertReceiver]:
        """
        Get a specific alert receivers by name.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        alert_receiver = alerts_api.get_alert_receiver("email")

        ```
        # Arguments
            name: The name of the alert receiver to retrieve.
        # Returns
            `AlertReceiver`: The alert receiver object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()

        name = self._fix_receiver_name(name, _client._project_name)
        path_params = ["project", _client._project_id, "alerts", "receivers", name]
        headers = {"content-type": "application/json"}
        return alert_receiver.AlertReceiver.from_response_json(
            _client._send_request("GET", path_params, headers=headers)
        )

    @usage.method_logger
    def create_alert_receiver(
        self,
        name: str,
        email_configs: List[alert_receiver.EmailConfig] = None,
        slack_configs: List[alert_receiver.SlackConfig] = None,
        pagerduty_configs: List[alert_receiver.PagerDutyConfig] = None,
        webhook_configs: List[alert_receiver.WebhookConfig] = None,
    ) -> alert_receiver.AlertReceiver:
        """
        Create a new alert receiver.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        new_alert_receiver = alerts_api.create_alert_receiver(name="email", email_configs=[{"to": "email@mail.com"}])

        ```
        # Arguments
            name: The name of the alert receiver (e.g., email, webhook).
            email_configs: List of email configurations (optional).
            slack_configs: List of Slack configurations (optional).
            pagerduty_configs: List of PagerDuty configurations (optional).
            webhook_configs: List of webhook configurations (optional).
        # Returns
            `AlertReceiver`: The created alert receiver object.
        # Raises
            `ValueError`: If multiple configurations are provided.
            `ValueError`: If the global channel for the configuration is not configured.
            'TimeoutError': If the alert receiver creation times out.
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        configured_receivers = self._get_configured_receivers()
        if (
            email_configs is not None
            and configured_receivers["email_configured"] is False
        ):
            raise ValueError(
                "Email channel is not configured. Please contact an administrator."
            )
        if (
            slack_configs is not None
            and configured_receivers["slack_configured"] is False
        ):
            raise ValueError(
                "Slack channel is not configured. Please contact an administrator."
            )
        if (
            pagerduty_configs is not None
            and configured_receivers["pager_duty_configured"] is False
        ):
            raise ValueError(
                "PagerDuty channel is not configured. Please contact an administrator."
            )

        non_none_args = [
            arg
            for arg in (
                email_configs,
                slack_configs,
                pagerduty_configs,
                webhook_configs,
            )
            if arg is not None
        ]
        if len(non_none_args) > 1:
            raise ValueError(
                "Only one of email_configs, slack_configs, pagerduty_configs, or webhook_configs can be provided."
            )

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "alerts", "receivers"]
        headers = {"content-type": "application/json"}
        data = {
            "name": name,
            "emailConfigs": self._validate_receiver_config(
                email_configs, alert_receiver.EmailConfig
            ),
            "slackConfigs": self._validate_receiver_config(
                slack_configs, alert_receiver.SlackConfig
            ),
            "pagerDutyConfigs": self._validate_receiver_config(
                pagerduty_configs, alert_receiver.PagerDutyConfig
            ),
            "webhookConfigs": self._validate_receiver_config(
                webhook_configs, alert_receiver.WebhookConfig
            ),
        }
        query_params = {"defaultTemplate": True}

        _client._send_request(
            "POST",
            path_params,
            data=json.dumps(data),
            query_params=query_params,
            headers=headers,
        )
        return alerts_engine.AlertsEngine().await_receiver(name)

    @usage.method_logger
    def delete_alert(self, alert_id: int):
        """
        Delete an alert by ID.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        alerts_api.delete_alert(alert_id=1)

        ```
        # Arguments
            alert_id: The ID of the alert to delete.
        # Returns
            `None`
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "service", "alerts", alert_id]
        headers = {"content-type": "application/json"}
        _client._send_request("DELETE", path_params, headers=headers)
        self._log.info(f"Alert with ID {alert_id} deleted successfully.")

    @usage.method_logger
    def trigger_alert(
        self,
        receiver_name: str,
        title: str,
        summary: str,
        description: str,
        severity: _SEVERITY_ARG,
        status: str,
        name: str,
        generator_url: str = None,
        expire_after_sec: int = None,
    ):
        """
        Trigger an alert.
        ```python

        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        alerts_api.trigger_alert(receiver_name="email", title="Title", summary="Alert summary", description="Alert description", severity="info", status="script_finished", name="my_alert")

        ```
        # Arguments
            receiver_name: The receiver of the alert (e.g., email, webhook).
            summary: The summary of the alert.
            description: The description of the alert.
            severity: The severity of the alert (warning, critical, info).
            status: The status of the alert.
            name: The name of the alert.
            generator_url: The URL of the alert generator (optional).
            expire_after_sec: The time in seconds after which the alert should expire (optional).
        # Returns
            `None`
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if severity not in _SEVERITY:
            raise ValueError(f"Severity must be one of the following: {_SEVERITY}.")
        _client = client.get_instance()
        self._create_route_if_not_exist(receiver_name, status, severity)
        path_params = ["project", _client._project_id, "alerts"]
        headers = {"content-type": "application/json"}
        data = {
            "alerts": [
                {
                    "labels": [
                        {"key": "alertname", "value": name},
                        {"key": "severity", "value": severity},
                        {"key": "status", "value": status},
                        _API_MATCH,
                    ],
                    "annotations": [
                        {"key": "title", "value": title},
                        {"key": "summary", "value": summary},
                        {"key": "description", "value": description},
                    ],
                    "startsAt": datetime.now().isoformat()
                    if expire_after_sec
                    else None,
                    "endsAt": (
                        datetime.now() + timedelta(seconds=expire_after_sec)
                    ).isoformat()
                    if expire_after_sec
                    else None,
                    "generatorURL": generator_url,
                }
            ]
        }
        _client._send_request(
            "POST", path_params, data=json.dumps(data), headers=headers
        )

    @usage.method_logger
    def get_triggered_alerts(
        self, active: bool = True, silenced: bool = False, inhibited: bool = False
    ) -> List[triggered_alert.TriggeredAlert]:
        """
        Get triggered alerts.
        ```python
        import hopsworks

        project = hopsworks.login()

        alerts_api = project.get_alerts_api()

        triggered_alerts = alerts_api.get_triggered_alerts()

        ```
        # Arguments
            active: Whether to include active alerts (default is True).
            silenced: Whether to include silenced alerts (default is False).
            inhibited: Whether to include inhibited alerts (default is False).
        # Returns
            `List[TriggeredAlert]`: The triggered alert objects.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "alerts"]
        query_params = {"active": active, "silenced": silenced, "inhibited": inhibited}
        headers = {"content-type": "application/json"}
        return triggered_alert.TriggeredAlert.from_response_json(
            _client._send_request(
                "GET", path_params, query_params=query_params, headers=headers
            )
        )

    def _get_configured_receivers(self):
        """
        Get configured alert receivers.
        :return: A list of configured alert receivers.
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "alerts", "receivers", "default"]
        headers = {"content-type": "application/json"}
        return humps.decamelize(
            _client._send_request("GET", path_params, headers=headers)
        )

    def _create_route_if_not_exist(
        self, receiver_name: str, status: str, severity: _SEVERITY_ARG
    ):
        """
        Create a route for the alert receiver.
        :return: None
        """
        _client = client.get_instance()
        if receiver_name is None:
            raise ValueError("Receiver name cannot be None.")

        receiver_name = self._fix_receiver_name(receiver_name, _client._project_name)
        # Check if the receiver exists
        receiver = self.get_alert_receiver(receiver_name)
        if receiver is None:
            raise ValueError(f"Receiver {receiver_name} does not exist.")

        # Only create a route if the receiver is a project receiver
        if receiver_name.startswith(f"{_client._project_name}__"):
            path_params = ["project", _client._project_id, "alerts", "routes"]
            headers = {"content-type": "application/json"}
            match = [
                {"key": "status", "value": status},
                {"key": "severity", "value": severity},
                _API_MATCH,
            ]
            data = {
                "receiver": receiver_name,
                "status": status,
                "groupBy": ["..."],  # aggregate by all possible labels
                "match": match,
                "continue": True,
            }

            try:
                _client._send_request(
                    "POST", path_params, data=json.dumps(data), headers=headers
                )
                alerts_engine.AlertsEngine().await_route(
                    receiver_name,
                    match,
                )
            except client.exceptions.RestAPIError as e:
                if (
                    e.response.status_code == 400
                    and e.response.json().get("errorCode", "") == 390002
                    and "already exists" in e.response.json().get("errorMsg", "")
                ):
                    self._log.info(
                        f"Route for receiver {receiver_name} and match {status} already exists."
                    )
                else:
                    raise e

    def _fix_receiver_name(self, name, project_name):
        """
        Fix the receiver name by adding the project name prefix if necessary.
        :param name: The name of the receiver.
        :param project_name: The name of the project.
        :return: The fixed receiver name.
        """
        if not name.startswith(f"{_GLOBAL_RECEIVER_PREFIX}__") and not name.startswith(
            f"{project_name}__"
        ):
            return f"{project_name}__{name}"
        return name

    def _validate_receiver_config(self, items, cls):
        if items is None or len(items) == 0:
            return None
        if not isinstance(items, list):
            raise ValueError(f"Expected list, got {type(items)}")
        normalized = []
        for item in items:
            if isinstance(item, cls):
                normalized.append(item.to_dict())
            elif isinstance(item, dict):
                json_decamelized = humps.decamelize(item)
                normalized.append(cls(**json_decamelized).to_dict())
            else:
                raise ValueError(f"Invalid item type: {type(item)} for {cls}")
        return normalized



#
#  Copyright 2021. Logical Clocks AB
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Union

import humps
from hsfs import util
from hsfs.constructor import prepared_statement_parameter


class ServingPreparedStatement:
    def __init__(
        self,
        feature_group_id: Optional[int] = None,
        prepared_statement_index: Optional[int] = None,
        prepared_statement_parameters: Optional[
            List[prepared_statement_parameter.PreparedStatementParameter]
        ] = None,
        query_online: Optional[str] = None,
        prefix: Optional[str] = None,
        type: Optional[str] = None,
        items: Optional[List[Dict[str, Any]]] = None,
        count: Optional[int] = None,
        href: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._feature_group_id = feature_group_id
        self._prepared_statement_index = prepared_statement_index
        # use setter to ensure that the parameters are sorted by index
        self.prepared_statement_parameters = prepared_statement_parameters
        self._query_online = query_online
        self._prefix = prefix

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> List[ServingPreparedStatement]:
        json_decamelized = humps.decamelize(json_dict)
        if json_decamelized["count"] == 0:
            return []
        else:
            return [cls(**pstm_dto) for pstm_dto in json_decamelized["items"]]

    def update_from_response_json(
        self, json_dict: Dict[str, Any]
    ) -> ServingPreparedStatement:
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "preparedStatementIndex": self._prepared_statement_index,
            "preparedStatementParameters": self._prepared_statement_parameters,
            "queryOnline": self._query_online,
        }

    def __repr__(self) -> str:
        repr_dict = humps.decamelize(self.to_dict())
        repr_dict["feature_group_id"] = self._feature_group_id
        repr_dict["prefix"] = self._prefix
        repr_dict["prepared_statement_parameters"] = [
            pstm_param.__repr__() for pstm_param in self._prepared_statement_parameters
        ]
        return json.dumps(repr_dict, sort_keys=True, indent=4)

    @property
    def feature_group_id(self) -> Optional[int]:
        return self._feature_group_id

    @property
    def prepared_statement_index(self) -> Optional[int]:
        return self._prepared_statement_index

    @property
    def prepared_statement_parameters(
        self,
    ) -> Optional[List[prepared_statement_parameter.PreparedStatementParameter]]:
        return self._prepared_statement_parameters

    @property
    def query_online(self) -> Optional[str]:
        return self._query_online

    @property
    def prefix(self) -> Optional[str]:
        return self._prefix

    @feature_group_id.setter
    def feature_group_id(self, feature_group_id: Optional[int]) -> None:
        self._feature_group_id = feature_group_id

    @prepared_statement_index.setter
    def prepared_statement_index(self, prepared_statement_index: Optional[int]) -> None:
        self._prepared_statement_index = prepared_statement_index

    @prepared_statement_parameters.setter
    def prepared_statement_parameters(
        self,
        prepared_statement_parameters: Union[
            List[prepared_statement_parameter.PreparedStatementParameter],
            List[Dict[str, Any]],
        ],
    ) -> None:
        if isinstance(prepared_statement_parameters[0], dict):
            prepared_statement_parameters = [
                prepared_statement_parameter.PreparedStatementParameter.from_response_json(
                    pstm_param
                )
                for pstm_param in prepared_statement_parameters
            ]

        self._prepared_statement_parameters = sorted(
            prepared_statement_parameters, key=lambda x: x.index
        )

    @query_online.setter
    def query_online(self, query_online: Optional[str]) -> None:
        self._query_online = query_online

    @prefix.setter
    def prefix(self, prefix: Optional[str]) -> None:
        self._prefix = prefix



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import inspect

from hsml import constants


class TestConstants:
    # NOTE
    # This class contains validations for constants and enum values.
    # The purpose of this class is to ensure that (1) we don't make undesired changes to contant values
    # that might break things somewhere else, and (2) we remember to update the pytests accordingly by
    # adding / removing / updating tests.

    # This class includes the following validations:
    # - Number of possible values of an Enum (to check for added/removed values)
    # - Exact values of contants (to check for modified values)

    # MODEL

    def test_model_framework_constants(self):
        # Arrange
        model_frameworks = {
            "FRAMEWORK_TENSORFLOW": "TENSORFLOW",
            "FRAMEWORK_TORCH": "TORCH",
            "FRAMEWORK_PYTHON": "PYTHON",
            "FRAMEWORK_SKLEARN": "SKLEARN",
            "FRAMEWORK_LLM": "LLM",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.MODEL,
            num_values=len(model_frameworks),
            expected_constants=model_frameworks,
            prefix="FRAMEWORK",
        )

    # MODEL_REGISTRY

    def test_model_registry_constants(self):
        # Arrange
        model_registry = {
            "HOPSFS_MOUNT_PREFIX": "/hopsfs/",
            "MODEL_FILES_DIR_NAME": "Files",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.MODEL_REGISTRY,
            num_values=len(model_registry),
            expected_constants=model_registry,
        )

    # MODEL_SERVING

    def test_model_serving_constants(self):
        # Arrange
        model_serving = {"MODELS_DATASET": "Models", "ARTIFACTS_DIR_NAME": "Artifacts"}

        # Assert
        self._check_added_modified_or_removed_values(
            constants.MODEL_SERVING,
            num_values=len(model_serving),
            expected_constants=model_serving,
        )

    # ARTIFACT_VERSION

    def test_artifact_version_constants(self):
        # Arrange
        artifact_versions = {"CREATE": "CREATE"}

        # Assert
        self._check_added_modified_or_removed_values(
            constants.ARTIFACT_VERSION,
            num_values=len(artifact_versions),
            expected_constants=artifact_versions,
        )

    # RESOURCES

    def test_resources_min_constants(self):
        # Arrange
        min_resources = {
            "MIN_NUM_INSTANCES": 1,
            "MIN_CORES": 0.2,
            "MIN_MEMORY": 32,
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.RESOURCES,
            num_values=len(min_resources),
            expected_constants=min_resources,
            prefix="MIN",
        )

    def test_resources_max_constants(self):
        # Arrange
        max_resources = {
            "MAX_CORES": 2,
            "MAX_MEMORY": 1024,
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.RESOURCES,
            num_values=len(max_resources),
            expected_constants=max_resources,
            prefix="MAX",
        )

    def test_resources_gpu_constants(self):
        # Assert
        assert hasattr(constants.RESOURCES, "GPUS")
        assert constants.RESOURCES.GPUS == 0

    # KAFKA_TOPIC

    def test_kafka_topic_names_constants(self):
        # Arrange
        kafka_topic_cons = {
            "NONE": "NONE",
            "CREATE": "CREATE",
            "NUM_REPLICAS": 1,
            "NUM_PARTITIONS": 1,
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.KAFKA_TOPIC,
            num_values=len(kafka_topic_cons),
            expected_constants=kafka_topic_cons,
        )

    # INFERENCE_LOGGER

    def test_inference_logger_constants(self):
        # Arrange
        if_modes = {
            "MODE_NONE": "NONE",
            "MODE_ALL": "ALL",
            "MODE_MODEL_INPUTS": "MODEL_INPUTS",
            "MODE_PREDICTIONS": "PREDICTIONS",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.INFERENCE_LOGGER,
            num_values=len(if_modes),
            expected_constants=if_modes,
            prefix="MODE",
        )

    # INFERENCE_BATCHER

    def test_inference_batcher_constants(self):
        # Arrange
        if_batcher = {"ENABLED": False}

        # Assert
        self._check_added_modified_or_removed_values(
            constants.INFERENCE_BATCHER,
            num_values=len(if_batcher),
            expected_constants=if_batcher,
        )

    # DEPLOYMENT

    def test_deployment_constants(self):
        # Arrange
        depl_actions = {"ACTION_START": "START", "ACTION_STOP": "STOP"}

        # Assert
        self._check_added_modified_or_removed_values(
            constants.DEPLOYMENT,
            num_values=len(depl_actions),
            expected_constants=depl_actions,
            prefix="ACTION",
        )

    # PREDICTOR

    def test_predictor_model_server_constants(self):
        # Arrange
        model_servers = {
            "MODEL_SERVER_PYTHON": "PYTHON",
            "MODEL_SERVER_TF_SERVING": "TENSORFLOW_SERVING",
            "MODEL_SERVER_VLLM": "VLLM",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.PREDICTOR,
            num_values=len(model_servers),
            expected_constants=model_servers,
            prefix="MODEL_SERVER",
        )

    def test_predictor_serving_tool_constants(self):
        # Arrange
        serving_tools = {
            "SERVING_TOOL_DEFAULT": "DEFAULT",
            "SERVING_TOOL_KSERVE": "KSERVE",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.PREDICTOR,
            num_values=len(serving_tools),
            expected_constants=serving_tools,
            prefix="SERVING_TOOL",
        )

    # PREDICTOR_STATE

    def test_predictor_state_status_constants(self):
        # Arrange
        predictor_states = {
            "STATUS_CREATING": "Creating",
            "STATUS_CREATED": "Created",
            "STATUS_STARTING": "Starting",
            "STATUS_FAILED": "Failed",
            "STATUS_RUNNING": "Running",
            "STATUS_IDLE": "Idle",
            "STATUS_UPDATING": "Updating",
            "STATUS_STOPPING": "Stopping",
            "STATUS_STOPPED": "Stopped",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.PREDICTOR_STATE,
            num_values=len(predictor_states),
            expected_constants=predictor_states,
            prefix="STATUS",
        )

    def test_predictor_state_condition_constants(self):
        # Arrange
        predictor_states = {
            "CONDITION_TYPE_STOPPED": "STOPPED",
            "CONDITION_TYPE_SCHEDULED": "SCHEDULED",
            "CONDITION_TYPE_INITIALIZED": "INITIALIZED",
            "CONDITION_TYPE_STARTED": "STARTED",
            "CONDITION_TYPE_READY": "READY",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.PREDICTOR_STATE,
            num_values=len(predictor_states),
            expected_constants=predictor_states,
            prefix="CONDITION",
        )

    # INFERENCE_ENDPOINTS

    def test_inference_endpoints_type_constants(self):
        # Arrange
        ie_types = {
            "ENDPOINT_TYPE_KUBE_CLUSTER": "KUBE_CLUSTER",
            "ENDPOINT_TYPE_LOAD_BALANCER": "LOAD_BALANCER",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.INFERENCE_ENDPOINTS,
            num_values=len(ie_types),
            expected_constants=ie_types,
            prefix="ENDPOINT_TYPE",
        )

    def test_inference_endpoints_port_constants(self):
        # Arrange
        ie_ports = {
            "PORT_NAME_HTTP": "HTTP",
            "PORT_NAME_HTTPS": "HTTPS",
            "PORT_NAME_STATUS_PORT": "STATUS",
            "PORT_NAME_TLS": "TLS",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.INFERENCE_ENDPOINTS,
            num_values=len(ie_ports),
            expected_constants=ie_ports,
            prefix="PORT_NAME",
        )

    def test_inference_endpoints_api_protocol_constants(self):
        # Arrange
        ie_api_protocols = {
            "API_PROTOCOL_REST": "REST",
            "API_PROTOCOL_GRPC": "GRPC",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.INFERENCE_ENDPOINTS,
            num_values=len(ie_api_protocols),
            expected_constants=ie_api_protocols,
            prefix="API_PROTOCOL",
        )

    # DEPLOYABLE_COMPONENT

    def test_inference_endpoints_deployable_component_constants(self):
        # Arrange
        depl_components = {
            "PREDICTOR": "predictor",
            "TRANSFORMER": "transformer",
        }

        # Assert
        self._check_added_modified_or_removed_values(
            constants.DEPLOYABLE_COMPONENT,
            num_values=len(depl_components),
            expected_constants=depl_components,
        )

    # Auxiliary methods

    def _check_added_modified_or_removed_values(
        self, cls, num_values, expected_constants=None, prefix=None
    ):
        cname = cls.__name__ + ("." + prefix if prefix is not None else "")
        const_dict = self._get_contants_name_value_dict(cls, prefix=prefix)
        # exact constants
        if expected_constants is not None:
            # constant names
            added_cnames = const_dict.keys() - expected_constants.keys()
            removed_cnames = expected_constants.keys() - const_dict.keys()

            assert len(added_cnames) == 0, (
                f"One or more constants were added under {cname} with names {added_cnames}. "
                + "If it was intentional, please add/remove/update tests accordingly (not only in this file, "
                + "but wherever it corresponds)."
            )

            assert len(removed_cnames) == 0, (
                f"One or more constants were removed under {cname} with names {removed_cnames}. "
                + "If it was intentional, please add/remove/update tests accordingly (not only in this file, "
                + "but wherever it corresponds)."
            )

            assert const_dict.keys() == expected_constants.keys(), (
                f"One or more constants under {cname} were modified from {removed_cnames} to {added_cnames}. "
                + "If it was intentional, please add/remove/update tests accordingly (not only in this file, "
                + "but wherever it corresponds)."
            )

            # constant values
            for cname, cvalue in expected_constants.items():
                full_cname = f"{cls.__name__}.{cname}"
                assert cvalue == const_dict[cname], (
                    f"The constant {full_cname} was modified from {cvalue} to {const_dict[cname]}. "
                    + "If it was intentional, please add/remove/update tests accordingly (not only in this file, "
                    + "but wherever it corresponds)."
                )
        else:
            # number of values
            assert len(const_dict) == num_values, (
                f"A constant was added/removed under {cname}. If it was intentional, please "
                + "add/remove/update tests accordingly (not only in this file, but wherever it corresponds)."
            )

    def _get_contants_name_value_dict(self, cls, prefix=None) -> dict:
        const_dict = dict()
        for m in inspect.getmembers(cls, lambda m: not (inspect.isroutine(m))):
            n = m[0]  # name
            if (prefix is not None and n.startswith(prefix)) or (
                prefix is None and not (n.startswith("__") and n.endswith("__"))
            ):
                const_dict[n] = m[1]  # value
        return const_dict



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import feature_group, training_dataset_feature
from hsfs.transformation_function import TransformationType


class TestTrainingDatasetFeature:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["training_dataset_feature"]["get"]["response"]

        # Act
        td_feature = training_dataset_feature.TrainingDatasetFeature.from_response_json(
            json
        )

        # Assert
        assert td_feature.name == "test_name"
        assert td_feature.type == "test_type"
        assert td_feature.index == "test_index"
        assert isinstance(td_feature._feature_group, feature_group.FeatureGroup)
        assert (
            td_feature._feature_group_feature_name == "test_feature_group_feature_name"
        )
        assert td_feature.label == "test_label"

    def test_from_response_json_on_demand_transformation(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["training_dataset_feature"]["get_transformations"][
            "response"
        ]

        # Act
        td_feature = training_dataset_feature.TrainingDatasetFeature.from_response_json(
            json
        )

        # Assert
        assert td_feature.name == "test_name"
        assert td_feature.type == "test_type"
        assert td_feature.index == "test_index"
        assert (
            td_feature.on_demand_transformation_function.hopsworks_udf.function_name
            == "add_one_fs"
        )

        assert (
            td_feature.on_demand_transformation_function.hopsworks_udf._function_source
            == "\n@udf(float)\ndef add_one_fs(data1 : pd.Series):\n    return data1 + 1\n"
        )
        assert (
            td_feature.on_demand_transformation_function.transformation_type
            == TransformationType.ON_DEMAND
        )
        assert isinstance(td_feature._feature_group, feature_group.FeatureGroup)
        assert (
            td_feature._feature_group_feature_name == "test_feature_group_feature_name"
        )
        assert td_feature.label == "test_label"

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["training_dataset_feature"]["get_basic_info"][
            "response"
        ]

        # Act
        td_feature = training_dataset_feature.TrainingDatasetFeature.from_response_json(
            json
        )

        # Assert
        assert td_feature.name == "test_name"
        assert td_feature.type is None
        assert td_feature.index is None
        assert td_feature._feature_group is None
        assert td_feature._feature_group_feature_name is None
        assert td_feature.label is False



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable, Dict, Literal, Optional, Tuple, Union

from hopsworks_common import client
from hopsworks_common.core.constants import (
    HAS_AVRO,
    HAS_CONFLUENT_KAFKA,
    HAS_FAST_AVRO,
    HAS_NUMPY,
    HAS_PANDAS,
    avro_not_installed_message,
)
from hopsworks_common.decorators import uses_confluent_kafka
from hsfs.core import online_ingestion, online_ingestion_api, storage_connector_api
from tqdm import tqdm


if HAS_NUMPY:
    import numpy as np

if HAS_PANDAS:
    import pandas as pd

if HAS_CONFLUENT_KAFKA:
    from confluent_kafka import Consumer, KafkaError, Producer, TopicPartition

if HAS_FAST_AVRO:
    from fastavro import schemaless_writer
    from fastavro.schema import parse_schema
elif HAS_AVRO:
    import avro.io
    import avro.schema


if TYPE_CHECKING:
    from hsfs.feature_group import ExternalFeatureGroup, FeatureGroup


@uses_confluent_kafka
def init_kafka_consumer(
    feature_store_id: int,
    offline_write_options: Dict[str, Any],
) -> Consumer:
    # setup kafka consumer
    consumer_config = get_kafka_config(feature_store_id, offline_write_options)
    if "group.id" not in consumer_config:
        consumer_config["group.id"] = "hsfs_consumer_group"

    return Consumer(consumer_config)


def init_kafka_resources(
    feature_group: Union[FeatureGroup, ExternalFeatureGroup],
    offline_write_options: Dict[str, Any],
    num_entries: Optional[int] = None,
) -> Tuple[
    Producer, Dict[str, bytes], Dict[str, Callable[..., bytes]], Callable[..., bytes] :
]:
    # this function is a caching wrapper around _init_kafka_resources
    if feature_group._multi_part_insert and feature_group._kafka_producer:
        return (
            feature_group._kafka_producer,
            feature_group._kafka_headers,
            feature_group._feature_writers,
            feature_group._writer,
        )
    producer, headers, feature_writers, writer = _init_kafka_resources(
        feature_group, offline_write_options, num_entries
    )
    if feature_group._multi_part_insert:
        feature_group._kafka_producer = producer
        feature_group._kafka_headers = headers
        feature_group._feature_writers = feature_writers
        feature_group._writer = writer
    return producer, headers, feature_writers, writer


def _init_kafka_resources(
    feature_group: Union[FeatureGroup, ExternalFeatureGroup],
    offline_write_options: Dict[str, Any],
    num_entries: Optional[int] = None,
) -> Tuple[
    Producer, Dict[str, bytes], Dict[str, Callable[..., bytes]], Callable[..., bytes] :
]:
    # setup kafka producer
    producer = init_kafka_producer(
        feature_group.feature_store_id, offline_write_options
    )
    # setup headers
    headers = get_headers(feature_group, num_entries)
    # setup writers
    feature_writers, writer = get_writer_function(feature_group)

    return producer, headers, feature_writers, writer


def get_writer_function(
    feature_group: Union[FeatureGroup, ExternalFeatureGroup],
) -> Tuple[Dict[str, Callable[..., bytes]], Callable[..., bytes]]:
    # setup complex feature writers
    feature_writers = {
        feature: get_encoder_func(feature_group._get_feature_avro_schema(feature))
        for feature in feature_group.get_complex_features()
    }
    # setup row writer function
    writer = get_encoder_func(feature_group._get_encoded_avro_schema())
    return (feature_writers, writer)


def get_headers(
    feature_group: Union[FeatureGroup, ExternalFeatureGroup],
    num_entries: Optional[int] = None,
) -> Dict[str, bytes]:
    # custom headers for hopsworks onlineFS
    headers = {
        "projectId": str(feature_group.feature_store.project_id).encode("utf8"),
        "featureGroupId": str(feature_group._id).encode("utf8"),
        "subjectId": str(feature_group.subject["id"]).encode("utf8"),
    }

    if feature_group.online_enabled:
        # setup online ingestion id
        online_ingestion_instance = (
            online_ingestion_api.OnlineIngestionApi().create_online_ingestion(
                feature_group, online_ingestion.OnlineIngestion(num_entries=num_entries)
            )
        )
        headers["onlineIngestionId"] = str(online_ingestion_instance.id).encode("utf8")

    return headers


@uses_confluent_kafka
def init_kafka_producer(
    feature_store_id: int,
    offline_write_options: Dict[str, Any],
) -> Producer:
    # setup kafka producer
    return Producer(get_kafka_config(feature_store_id, offline_write_options))


@uses_confluent_kafka
def kafka_get_offsets(
    topic_name: str,
    feature_store_id: int,
    offline_write_options: Dict[str, Any],
    high: bool,
) -> str:
    consumer = init_kafka_consumer(feature_store_id, offline_write_options)
    topics = consumer.list_topics(
        timeout=offline_write_options.get("kafka_timeout", 6)
    ).topics
    if topic_name in topics.keys():
        # topic exists
        offsets = ""
        tuple_value = int(high)
        for partition_metadata in topics.get(topic_name).partitions.values():
            partition = TopicPartition(
                topic=topic_name, partition=partition_metadata.id
            )
            offsets += f",{partition_metadata.id}:{consumer.get_watermark_offsets(partition)[tuple_value]}"
        consumer.close()

        return f"{topic_name + offsets}"
    return ""


def kafka_produce(
    producer: Producer,
    key: str,
    encoded_row: bytes,
    topic_name: str,
    headers: Dict[str, bytes],
    acked: callable,
    debug_kafka: bool = False,
) -> None:
    while True:
        # if BufferError is thrown, we can be sure, message hasn't been send so we retry
        try:
            # produce
            producer.produce(
                topic=topic_name,
                key=key,
                value=encoded_row,
                callback=acked,
                headers=headers,
            )

            # Trigger internal callbacks to empty op queue
            producer.poll(0)
            break
        except BufferError as e:
            if debug_kafka:
                print("Caught: {}".format(e))
            # backoff for 1 second
            producer.poll(1)


def encode_complex_features(
    feature_writers: Dict[str, callable], row: Dict[str, Any]
) -> Dict[str, Any]:
    for feature_name, writer in feature_writers.items():
        with BytesIO() as outf:
            writer(row[feature_name], outf)
            row[feature_name] = outf.getvalue()
    return row


def get_encoder_func(writer_schema: str) -> callable:
    if HAS_FAST_AVRO:
        schema = json.loads(writer_schema)
        parsed_schema = parse_schema(schema)
        return lambda record, outf: schemaless_writer(outf, parsed_schema, record)

    if not HAS_AVRO:
        raise ModuleNotFoundError(avro_not_installed_message)

    parsed_schema = avro.schema.parse(writer_schema)
    writer = avro.io.DatumWriter(parsed_schema)
    return lambda record, outf: writer.write(record, avro.io.BinaryEncoder(outf))


def encode_row(complex_feature_writers, writer, row):
    # transform special data types
    # here we might need to handle also timestamps and other complex types
    # possible optimizaiton: make it based on type so we don't need to loop over
    # all keys in the row
    if isinstance(row, dict):
        for k in row.keys():
            # for avro to be able to serialize them, they need to be python data types
            if HAS_NUMPY and isinstance(row[k], np.ndarray):
                row[k] = row[k].tolist()
            if HAS_PANDAS and isinstance(row[k], pd.Timestamp):
                row[k] = row[k].to_pydatetime()
            if isinstance(row[k], datetime) and row[k].tzinfo is None:
                row[k] = row[k].replace(tzinfo=timezone.utc)
            if HAS_PANDAS and isinstance(row[k], pd._libs.missing.NAType):
                row[k] = None
    # encode complex features
    row = encode_complex_features(complex_feature_writers, row)
    # encode feature row
    with BytesIO() as outf:
        writer(row, outf)
        encoded_row = outf.getvalue()
    return encoded_row


def get_kafka_config(
    feature_store_id: int,
    write_options: Optional[Dict[str, Any]] = None,
    engine: Literal["spark", "confluent"] = "confluent",
) -> Dict[str, Any]:
    if write_options is None:
        write_options = {}
    external = client._is_external() and not write_options.get("internal_kafka", False)

    storage_connector = storage_connector_api.StorageConnectorApi().get_kafka_connector(
        feature_store_id, external
    )

    if engine == "spark":
        config = storage_connector.spark_options()
        config.update(write_options)
    elif engine == "confluent":
        config = storage_connector.confluent_options()
        config.update(write_options.get("kafka_producer_config", {}))
    return config


@uses_confluent_kafka
def build_ack_callback_and_optional_progress_bar(
    n_rows: int, is_multi_part_insert: bool, offline_write_options: Dict[str, Any]
) -> Tuple[Callable, Optional[tqdm]]:
    if not is_multi_part_insert:
        progress_bar = tqdm(
            total=n_rows,
            bar_format="{desc}: {percentage:.2f}% |{bar}| Rows {n_fmt}/{total_fmt} | "
            "Elapsed Time: {elapsed} | Remaining Time: {remaining}",
            desc="Uploading Dataframe",
            mininterval=1,
        )
    else:
        progress_bar = None

    def acked(err: Exception, msg: Any) -> None:
        if err is not None:
            if offline_write_options.get("debug_kafka", False):
                print("Failed to deliver message: %s: %s" % (str(msg), str(err)))
            if err.code() in [
                KafkaError.TOPIC_AUTHORIZATION_FAILED,
                KafkaError._MSG_TIMED_OUT,
            ]:
                progress_bar.colour = "RED"
                raise err  # Stop producing and show error
        # update progress bar for each msg
        if not is_multi_part_insert:
            progress_bar.update()

    return acked, progress_bar



import random
import string
from unittest import mock

import pandas as pd
import pytest
from hopsworks_common.core.constants import HAS_POLARS
from hsfs import engine, feature_group
from hsfs.core.schema_validation import (
    DataFrameValidator,
    PandasValidator,
    PolarsValidator,
    PySparkValidator,
)
from hsfs.engine import spark
from hsfs.feature import Feature


if HAS_POLARS:
    import polars as pl


@pytest.fixture
def pandas_df():
    return pd.DataFrame(
        {
            "primary_key": [1, 2, 3],
            "event_time": ["2021-01-01", "2021-01-02", "2021-01-03"],
            "string_col": [
                (
                    "".join(
                        random.choice(string.ascii_letters)
                        for _ in range(random.randint(1, 100))
                    )
                )
                for _ in range(3)
            ],
        },
        dtype=pd.StringDtype(),
    )


@pytest.fixture
@pytest.mark.skipif(not HAS_POLARS, reason="polars not installed")
def polars_df():
    return pl.DataFrame(
        {
            "primary_key": [1, 2, 3],
            "event_time": ["2021-01-01", "2021-01-02", "2021-01-03"],
            "string_col": [
                (
                    "".join(
                        random.choice(string.ascii_letters)
                        for _ in range(random.randint(1, 100))
                    )
                )
                for _ in range(3)
            ],
        }
    )


@pytest.fixture
def spark_df():
    import random
    import string

    from pyspark.sql import Row

    spark_engine = spark.Engine()
    string_values = [
        "".join(
            random.choice(string.ascii_letters) for _ in range(random.randint(1, 100))
        )
        for _ in range(3)
    ]

    data = [
        Row(primary_key=1, event_time="2021-01-01", string_col=string_values[0]),
        Row(primary_key=2, event_time="2021-01-02", string_col=string_values[1]),
        Row(primary_key=3, event_time="2021-01-03", string_col=string_values[2]),
    ]

    return spark_engine._spark_session.createDataFrame(data)


@pytest.fixture
def feature_group_data():
    with mock.patch("hopsworks_common.client.get_instance"):
        engine.init("python")
    fg = feature_group.FeatureGroup(
        name="test_fg",
        version=1,
        online_enabled=True,
        partition_key=[],
        featurestore_id=1,
        primary_key=["primary_key"],
        features=[
            Feature("primary_key", "int", online_type="varchar(10)"),
            Feature("event_time", "string", online_type="varchar(10)"),
            Feature("string_col", "string", online_type="varchar(200)"),
        ],
    )
    return fg


@pytest.fixture
def feature_group_created():
    with mock.patch("hopsworks_common.client.get_instance"):
        engine.init("python")
    fg_existing = feature_group.FeatureGroup(
        name="test_existing_fg",
        id=1,
        version=1,
        online_enabled=True,
        partition_key=[],
        featurestore_id=1,
        primary_key=["primary_key"],
        features=[
            Feature("primary_key", "int", online_type="varchar(10)"),
            Feature("event_time", "string", online_type="varchar(10)"),
            Feature("string_col", "string", online_type="varchar(100)"),
        ],
    )
    return fg_existing


# Base class for common test behavior
class BaseDataFrameTest:
    def test_primary_key_missing(self, df, feature_group_data):
        KEY = "missing_key"
        feature_group_data.primary_key = [KEY]
        with pytest.raises(
            ValueError, match=f"Primary key column {KEY} is missing in input dataframe"
        ):
            DataFrameValidator().validate_schema(
                feature_group_data, df, feature_group_data.features
            )

    def test_feature_group_with_features_not_created(self, df, feature_group_data):
        # test with feature group with explicit features
        df_features = DataFrameValidator().validate_schema(
            feature_group_data, df, feature_group_data.features
        )
        # assert that the online type of the string_col feature is same as explcitly set in the feature group
        assert df_features[2].online_type == "varchar(200)"

    def test_primary_key_null(self, df, feature_group_data):
        modified_df = self._modify_row(df, 0, primary_key=None)
        with pytest.raises(
            ValueError, match="Primary key column primary_key contains null values"
        ):
            DataFrameValidator().validate_schema(
                feature_group_data, modified_df, feature_group_data.features
            )

    def test_string_length_exceeded(self, df, feature_group_created, mocker):
        modified_df = self._modify_row(df, 0, string_col="a" * 101)
        with pytest.raises(ValueError, match="String length exceeded"):
            DataFrameValidator().validate_schema(
                feature_group_created, modified_df, feature_group_created.features
            )

    def test_fg_features_string_length_exceeded(self, df, feature_group_data):
        modified_df = self._modify_row(df, 0, string_col="a" * 301)
        with pytest.raises(ValueError, match="String length exceeded"):
            DataFrameValidator().validate_schema(
                feature_group_data, modified_df, feature_group_data.features
            )

    def test_feature_group_created(self, df, feature_group_created, mocker):
        modified_df = self._modify_row(df, 0, string_col="a" * 101)
        with pytest.raises(ValueError, match="String length exceeded"):
            DataFrameValidator().validate_schema(
                feature_group_created, modified_df, feature_group_created.features
            )

    def test_feature_group_not_created(self, df, feature_group_data):
        # test with non existing feature group with no explicit features
        # arrange
        modified_df = self._modify_row(df, 0, string_col="a" * 101)
        initial_features = [
            Feature("primary_key", "int"),
            Feature("event_time", "string"),
            Feature("string_col", "string"),
        ]
        feature_group_data.features = []
        df_features = DataFrameValidator().validate_schema(
            feature_group_data, modified_df, initial_features
        )
        assert df_features[2].online_type == "varchar(200)"

    def test_pk_null_string_length_exceeded(self, df, feature_group_data):
        modified_df = self._modify_row(df, 0, primary_key=None, string_col="a" * 101)
        with pytest.raises(
            ValueError, match="One or more schema validation errors found"
        ):
            DataFrameValidator().validate_schema(
                feature_group_data, modified_df, feature_group_data.features
            )

    def test_should_not_update_nonvarchar(self, df, feature_group_data):
        # Test that the validator does not update the online type of a non-varchar column
        # arrange
        # set string_col feature online type to text
        feature_group_data.features[2].online_type = "text"
        modified_df = self._modify_row(df, 0, string_col="b" * 1001)
        # act
        df_features = DataFrameValidator().validate_schema(
            feature_group_data, modified_df, feature_group_data.features
        )
        assert df_features == feature_group_data.features

    def test_offline_fg(self, df, feature_group_data, caplog):
        # test that offline fg are skipped for validations
        # arrange
        feature_group_data.online_enabled = False
        modified_df = self._modify_row(df, 0, primary_key=None, string_col="a" * 101)
        # act
        df_features = DataFrameValidator().validate_schema(
            feature_group_data, modified_df, feature_group_data.features
        )
        # assert raises warning
        assert "Feature group is not online enabled" in caplog.text
        # assert no changes were made
        assert df_features == feature_group_data.features

    def test_embedding_feature_group(self, df, feature_group_data, caplog):
        # Test that the embedding fg is skipped validation
        # arrange
        modified_df = self._modify_row(df, 0, primary_key=None, string_col="a" * 101)
        feature_group_data.embedding_index = "string_col"
        # act
        df_features = DataFrameValidator().validate_schema(
            feature_group_data, modified_df, feature_group_data.features
        )
        # assert raises warning
        assert "Feature group is embedding" in caplog.text
        assert df_features == feature_group_data.features


class TestPandasDataframe(BaseDataFrameTest):
    @pytest.fixture
    def df(self, pandas_df):
        return pandas_df

    def _modify_row(self, df, row_idx, **kwargs):
        """Helper method to modify a specific row in a Pandas DataFrame with custom index."""
        # Get the index value at the given position
        idx_value = df.index[row_idx]
        # Apply modifications for each column
        for column, value in kwargs.items():
            df.loc[idx_value, column] = value

        return df

    def test_get_validator_pandas(self, df):
        validator = DataFrameValidator.get_validator(df)
        assert isinstance(validator, PandasValidator)

    def test_string_column_detection(self):
        df = pd.DataFrame(
            {
                "pk": [1, 2, 3, 4, 5],
                "obj": [[1, 2], ["a", "b"], {"key": "value"}, None, [5, 6]],
                "string1": pd.Series(
                    ["hello", "world", "python", None, "test"], dtype="object"
                ),
                "string2": pd.Series(
                    ["dasd", "dfsadsfa", "dsadwasd", None, "tedfdsfdgst"],
                    dtype="string",
                ),
                "mixed": ["hello", 2.0, pd.Timestamp.now(), None, pd.Timestamp.now()],
            }
        )
        # act
        string_cols = PandasValidator.get_string_columns(df)
        # validate that only 'val' is detected as a string column
        assert ["string1", "string2"] == string_cols


@pytest.mark.skipif(not HAS_POLARS, reason="polars not installed")
class TestPolarsDataframe(BaseDataFrameTest):
    @pytest.fixture
    def df(self, polars_df):
        return polars_df

    def _modify_row(self, df, row_idx, **kwargs):
        """Helper method to modify a specific row in a Polars DataFrame."""
        expressions = []

        for column, value in kwargs.items():
            expressions.append(
                pl.when(pl.col("index") == row_idx)
                .then(pl.lit(value))
                .otherwise(pl.col(column))
                .alias(column)
            )

        if not expressions:
            return df

        return df.with_row_index().with_columns(expressions).drop("index")

    def test_get_validator_polars(self, df):
        validator = DataFrameValidator.get_validator(df)
        assert isinstance(validator, PolarsValidator)


class TestSparkDataframe(BaseDataFrameTest):
    @pytest.fixture
    def df(self, spark_df):
        return spark_df

    def _modify_row(self, df, row_idx, **kwargs):
        """Helper method to modify a specific row in a Spark DataFrame."""
        from pyspark.sql.functions import col, lit, row_number, when
        from pyspark.sql.window import Window

        # Create a window spec for row numbering
        window_spec = Window.orderBy(
            "primary_key"
        )  # Change this to an appropriate ordering column

        # Add row numbers
        df_with_row_nr = df.withColumn("row_nr", row_number().over(window_spec))

        # Apply modifications for each column
        result_df = df_with_row_nr
        for column, value in kwargs.items():
            result_df = result_df.withColumn(
                column,
                when(col("row_nr") == row_idx + 1, lit(value)).otherwise(col(column)),
            )

        # Remove the row_nr column
        return result_df.drop("row_nr")

    def test_get_validator_spark(self, spark_df):
        validator = DataFrameValidator.get_validator(spark_df)
        assert isinstance(validator, PySparkValidator)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import logging
import time
from datetime import datetime
from typing import List, Optional

from hopsworks_common import alert_receiver, alert_route, client


class AlertsEngine:
    def __init__(self):
        self._log = logging.getLogger(__name__)

    def await_receiver(self, receiver_name, timeout: Optional[float] = 120):
        """
        Waits for a receiver to be created. This is useful when creating a new receiver
        and you want to ensure that it is available before proceeding with other operations.
        :param receiver_name: The name of the receiver to wait for.
        :type receiver_name: str
        :param timeout: The maximum time to wait for the receiver to be created, in seconds.
        :type timeout: float
        :raises TimeoutError: If the receiver is not created within the specified timeout.
        :return: The created receiver.
        :rtype: AlertReceiver
        """
        if receiver_name is None:
            raise ValueError("Receiver name cannot be None.")

        start_time = datetime.now()

        def passed():
            return (datetime.now() - start_time).total_seconds()

        receiver = self._get_receiver(receiver_name)
        self._log.info("Waiting for receiver to be created.")
        while receiver is None:
            if timeout is not None and passed() > timeout:
                raise TimeoutError(
                    f"Receiver {receiver_name} was not created in {timeout} seconds."
                )
            time.sleep(5)
            receiver = self._get_receiver(receiver_name)

        self._log.info("Receiver created.")
        return receiver

    def _get_receiver(self, name):
        _client = client.get_instance()
        # should only wait for project receivers
        if not name.startswith(f"{_client._project_name}__"):
            name = f"{_client._project_name}__{name}"
        path_params = ["project", _client._project_id, "alerts", "receivers"]
        query_params = {"expand": True}
        headers = {"content-type": "application/json"}
        receivers = alert_receiver.AlertReceiver.from_response_json(
            _client._send_request(
                "GET", path_params, query_params=query_params, headers=headers
            )
        )
        if receivers is not None:
            for receiver in receivers:
                if receiver.name == name:
                    return receiver
        return None

    def await_route(
        self, receiver_name: str, match: List[dict], timeout: Optional[float] = 120
    ):
        """
        Waits for a route to be created. This is useful when creating a new route
        and you want to ensure that it is available before proceeding with other operations.
        :param receiver_name: The name of the receiver to wait for.
        :match: The match criteria for the route.
        :param timeout: The maximum time to wait for the route to be created, in seconds.
        :type timeout: float
        :raises TimeoutError: If the route is not created within the specified timeout.
        :return: The created route.
        :rtype: AlertRoute
        """
        if receiver_name is None:
            raise ValueError("Receiver name cannot be None.")

        start_time = datetime.now()

        def passed():
            return (datetime.now() - start_time).total_seconds()

        route = self._get_route(receiver_name, match)
        self._log.info("Waiting for route to be created.")
        while route is None:
            if timeout is not None and passed() > timeout:
                raise TimeoutError(
                    f"Route for receiver {receiver_name} was not created in {timeout} seconds."
                )
            time.sleep(5)
            route = self._get_route(receiver_name, match)

        self._log.info("Route created.")
        return route

    def _get_route(self, name: str, match: List[dict]):
        _client = client.get_instance()
        # should only wait for project receivers
        if not name.startswith(f"{_client._project_name}__"):
            name = f"{_client._project_name}__{name}"
        path_params = ["project", _client._project_id, "alerts", "routes"]
        query_params = {"loaded": True}  # we want to get the loaded routes
        headers = {"content-type": "application/json"}
        routes = alert_route.AlertRoute.from_response_json(
            _client._send_request(
                "GET", path_params, query_params=query_params, headers=headers
            )
        )
        if routes is not None:
            for route in routes:
                if route._receiver == name and all(
                    item["key"] in route._match
                    and route._match[item["key"]] == item["value"]
                    for item in match
                ):
                    return route
        return None



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
import tempfile
import time
import uuid
from typing import Dict, List, Union

from hopsworks_common.client.exceptions import ModelServingException, RestAPIError
from hopsworks_common.client.istio.utils.infer_type import InferInput
from hopsworks_common.constants import (
    DEPLOYMENT,
    MODEL_SERVING,
    PREDICTOR,
    PREDICTOR_STATE,
)
from hopsworks_common.constants import INFERENCE_ENDPOINTS as IE
from hopsworks_common.core import dataset_api
from hsml.core import serving_api
from hsml.engine import local_engine
from tqdm.auto import tqdm


class ServingEngine:
    START_STEPS = [
        PREDICTOR_STATE.CONDITION_TYPE_STOPPED,
        PREDICTOR_STATE.CONDITION_TYPE_SCHEDULED,
        PREDICTOR_STATE.CONDITION_TYPE_INITIALIZED,
        PREDICTOR_STATE.CONDITION_TYPE_STARTED,
        PREDICTOR_STATE.CONDITION_TYPE_READY,
    ]
    STOP_STEPS = [
        PREDICTOR_STATE.CONDITION_TYPE_SCHEDULED,
        PREDICTOR_STATE.CONDITION_TYPE_STOPPED,
    ]

    def __init__(self):
        self._serving_api = serving_api.ServingApi()
        self._dataset_api = dataset_api.DatasetApi()

        self._engine = local_engine.LocalEngine()

    def _poll_deployment_status(
        self, deployment_instance, status: str, await_status: int, update_progress=None
    ):
        if await_status > 0:
            sleep_seconds = 5
            for _ in range(int(await_status / sleep_seconds)):
                time.sleep(sleep_seconds)
                state = deployment_instance.get_state()
                num_instances = self._get_available_instances(state)
                if update_progress is not None:
                    update_progress(state, num_instances)
                if state.status == status:
                    return state  # deployment reached desired status
                elif (
                    status == PREDICTOR_STATE.STATUS_RUNNING
                    and state.status == PREDICTOR_STATE.STATUS_FAILED
                ):
                    error_msg = state.condition.reason
                    if (
                        state.condition.type
                        == PREDICTOR_STATE.CONDITION_TYPE_INITIALIZED
                        or state.condition.type
                        == PREDICTOR_STATE.CONDITION_TYPE_STARTED
                    ):
                        component = (
                            "transformer"
                            if "transformer" in state.condition.reason
                            else "predictor"
                        )
                        error_msg += (
                            ". Please, check the server logs using `.get_logs(component='"
                            + component
                            + "')`"
                        )
                    raise ModelServingException(error_msg)
            raise ModelServingException(
                "Deployment has not reached the desired status within the expected awaiting time. Check the current status by using `.get_state()`, "
                + "explore the server logs using `.get_logs()` or set a higher value for await_"
                + status.lower()
            )

    def start(self, deployment_instance, await_status: int) -> bool:
        (done, state) = self._check_status(
            deployment_instance, PREDICTOR_STATE.STATUS_RUNNING
        )

        if not done:
            min_instances = self._get_min_starting_instances(deployment_instance)
            num_steps = (len(self.START_STEPS) - 1) + min_instances
            if deployment_instance._predictor._state.condition is None:
                num_steps = min_instances  # backward compatibility
            pbar = tqdm(total=num_steps)
            pbar.set_description("Creating deployment")

            # set progress function
            def update_progress(state, num_instances):
                (progress, desc) = self._get_starting_progress(
                    pbar.n, state, num_instances
                )
                pbar.update(progress)
                if desc is not None:
                    pbar.set_description(desc)

            try:
                update_progress(state, num_instances=0)

                if state.status == PREDICTOR_STATE.STATUS_CREATING:
                    state = self._poll_deployment_status(  # wait for preparation
                        deployment_instance,
                        PREDICTOR_STATE.STATUS_CREATED,
                        await_status,
                        update_progress,
                    )

                self._serving_api.post(
                    deployment_instance, DEPLOYMENT.ACTION_START
                )  # start deployment

                state = self._poll_deployment_status(  # wait for status
                    deployment_instance,
                    PREDICTOR_STATE.STATUS_RUNNING,
                    await_status,
                    update_progress,
                )
            except RestAPIError as re:
                self.stop(deployment_instance, await_status=0)
                raise re

        if state.status == PREDICTOR_STATE.STATUS_RUNNING:
            print("Start making predictions by using `.predict()`")

    def stop(self, deployment_instance, await_status: int) -> bool:
        (done, state) = self._check_status(
            deployment_instance, PREDICTOR_STATE.STATUS_STOPPED
        )
        if not done:
            num_instances = self._get_available_instances(state)
            num_steps = len(self.STOP_STEPS) + (
                deployment_instance.requested_instances
                if deployment_instance.requested_instances >= num_instances
                else num_instances
            )
            if deployment_instance._predictor._state.condition is None:
                # backward compatibility
                num_steps = self._get_min_starting_instances(deployment_instance)
            pbar = tqdm(total=num_steps)
            pbar.set_description("Preparing to stop deployment")

            # set progress function
            def update_progress(state, num_instances):
                (progress, desc) = self._get_stopping_progress(
                    pbar.total, pbar.n, state, num_instances
                )
                pbar.update(progress)
                if desc is not None:
                    pbar.set_description(desc)

            update_progress(state, num_instances)
            self._serving_api.post(
                deployment_instance, DEPLOYMENT.ACTION_STOP
            )  # stop deployment

            _ = self._poll_deployment_status(  # wait for status
                deployment_instance,
                PREDICTOR_STATE.STATUS_STOPPED,
                await_status,
                update_progress,
            )

        # free grpc channel
        deployment_instance._grpc_channel = None

    def _check_status(self, deployment_instance, desired_status):
        state = deployment_instance.get_state()
        if state is None:
            return (True, None)

        # desired status: running
        if desired_status == PREDICTOR_STATE.STATUS_RUNNING:
            if (
                state.status == PREDICTOR_STATE.STATUS_RUNNING
                or state.status == PREDICTOR_STATE.STATUS_IDLE
            ):
                print("Deployment is already running")
                return (True, state)
            if state.status == PREDICTOR_STATE.STATUS_STARTING:
                print("Deployment is already starting")
                return (True, state)
            if state.status == PREDICTOR_STATE.STATUS_UPDATING:
                print("Deployments is already running and updating")
                return (True, state)
            if state.status == PREDICTOR_STATE.STATUS_FAILED:
                print("Deployment is in failed state. " + state.condition.reason)
                return (True, state)
            if state.status == PREDICTOR_STATE.STATUS_STOPPING:
                raise ModelServingException(
                    "Deployment is stopping, please wait until it completely stops"
                )

        # desired status: stopped
        if desired_status == PREDICTOR_STATE.STATUS_STOPPED:
            if (
                state.status == PREDICTOR_STATE.STATUS_CREATING
                or state.status == PREDICTOR_STATE.STATUS_CREATED
                or state.status == PREDICTOR_STATE.STATUS_STOPPED
            ):
                print("Deployment is already stopped")
                return (True, state)
            if state.status == PREDICTOR_STATE.STATUS_STOPPING:
                print("Deployment is already stopping")
                return (True, state)

        return (False, state)

    def _get_starting_progress(self, current_step, state, num_instances):
        if state.condition is None:  # backward compatibility
            progress = num_instances - current_step
            if state.status == PREDICTOR_STATE.STATUS_RUNNING:
                return (progress, "Deployment is ready")
            return (progress, None if current_step == 0 else "Deployment is starting")

        step = self.START_STEPS.index(state.condition.type)
        if (
            state.condition.type == PREDICTOR_STATE.CONDITION_TYPE_STARTED
            or state.condition.type == PREDICTOR_STATE.CONDITION_TYPE_READY
        ):
            step += num_instances
        progress = step - current_step
        desc = None
        if state.condition.type != PREDICTOR_STATE.CONDITION_TYPE_STOPPED:
            desc = (
                state.condition.reason
                if state.status != PREDICTOR_STATE.STATUS_FAILED
                else "Deployment failed to start"
            )
        return (progress, desc)

    def _get_stopping_progress(self, total_steps, current_step, state, num_instances):
        if state.condition is None:  # backward compatibility
            progress = (total_steps - num_instances) - current_step
            if state.status == PREDICTOR_STATE.STATUS_STOPPED:
                return (progress, "Deployment is stopped")
            return (
                progress,
                None if total_steps == current_step else "Deployment is stopping",
            )

        step = 0
        if state.condition.type == PREDICTOR_STATE.CONDITION_TYPE_SCHEDULED:
            step = 1 if state.condition.status is None else 0
        elif state.condition.type == PREDICTOR_STATE.CONDITION_TYPE_STOPPED:
            num_instances = (total_steps - 2) - num_instances  # num stopped instances
            step = (
                (2 + num_instances)
                if (state.condition.status is None or state.condition.status)
                else 0
            )
        progress = step - current_step
        desc = None
        if (
            state.condition.type != PREDICTOR_STATE.CONDITION_TYPE_READY
            and state.status != PREDICTOR_STATE.STATUS_FAILED
        ):
            desc = (
                "Deployment is stopped"
                if state.status == PREDICTOR_STATE.STATUS_STOPPED
                else state.condition.reason
            )

        return (progress, desc)

    def _get_min_starting_instances(self, deployment_instance):
        min_start_instances = 1  # predictor
        if deployment_instance.transformer is not None:
            min_start_instances += 1  # transformer
        return (
            deployment_instance.requested_instances
            if deployment_instance.requested_instances >= min_start_instances
            else min_start_instances
        )

    def _get_available_instances(self, state):
        if state.status == PREDICTOR_STATE.STATUS_CREATING:
            return 0
        num_instances = state.available_predictor_instances
        if state.available_transformer_instances is not None:
            num_instances += state.available_transformer_instances
        return num_instances

    def _get_stopped_instances(self, available_instances, requested_instances):
        num_instances = requested_instances - available_instances
        return num_instances if num_instances >= 0 else 0

    def _download_files_from_hopsfs_recursive(
        self,
        from_hdfs_path: str,
        to_local_path: str,
        update_download_progress,
        n_dirs,
        n_files,
    ):
        """Download model files from a model path in hdfs, recursively"""

        for entry in self._dataset_api.list(from_hdfs_path, sort_by="NAME:desc")[
            "items"
        ]:
            path_attr = entry["attributes"]
            path = path_attr["path"]
            basename = os.path.basename(path)

            if path_attr.get("dir", False):
                # otherwise, make a recursive call for the folder
                if (
                    basename == MODEL_SERVING.ARTIFACTS_DIR_NAME
                ):  # TODO: Not needed anymore
                    continue  # skip Artifacts subfolder
                local_folder_path = os.path.join(to_local_path, basename)
                os.mkdir(local_folder_path)
                n_dirs, n_files = self._download_files_from_hopsfs_recursive(
                    from_hdfs_path=path,
                    to_local_path=local_folder_path,
                    update_download_progress=update_download_progress,
                    n_dirs=n_dirs,
                    n_files=n_files,
                )
                n_dirs += 1
                update_download_progress(n_dirs=n_dirs, n_files=n_files)
            else:
                # if it's a file, download it
                local_file_path = os.path.join(to_local_path, basename)
                self._engine.download(path, local_file_path)
                n_files += 1
                update_download_progress(n_dirs=n_dirs, n_files=n_files)

        return n_dirs, n_files

    def _download_files_from_hopsfs(
        self, from_hdfs_path: str, to_local_path: str, update_download_progress
    ):
        """Download files from a model path in hdfs."""

        n_dirs, n_files = self._download_files_from_hopsfs_recursive(
            from_hdfs_path=from_hdfs_path,
            to_local_path=to_local_path,
            update_download_progress=update_download_progress,
            n_dirs=0,
            n_files=0,
        )
        update_download_progress(n_dirs=n_dirs, n_files=n_files, done=True)

    def download_artifact_files(self, deployment_instance, local_path=None):
        if deployment_instance.id is None:
            raise ModelServingException(
                "Deployment is not created yet. To create the deployment use `.save()`"
            )
        if deployment_instance.artifact_version is None:
            # model artifacts are not created in non-k8s installations
            raise ModelServingException(
                "Model artifacts not supported in non-k8s installations. \
                 Download the model files by using `model.download()`"
            )

        if local_path is None:
            local_path = os.path.join(
                tempfile.gettempdir(),
                str(uuid.uuid4()),
                deployment_instance.model_name,
                str(deployment_instance.model_version),
                MODEL_SERVING.ARTIFACTS_DIR_NAME,
                str(deployment_instance.artifact_version),
            )
        os.makedirs(local_path, exist_ok=True)

        def update_download_progress(n_dirs, n_files, done=False):
            print(
                "Downloading artifact files (%s dirs, %s files)... %s"
                % (n_dirs, n_files, "DONE" if done else ""),
                end="\r",
            )

        try:
            from_hdfs_path = deployment_instance.artifact_files_path
            if from_hdfs_path.startswith("hdfs:/"):
                projects_index = from_hdfs_path.find("/Projects", 0)
                from_hdfs_path = from_hdfs_path[projects_index:]

            self._download_files_from_hopsfs(
                from_hdfs_path=from_hdfs_path,
                to_local_path=local_path,
                update_download_progress=update_download_progress,
            )
        except BaseException as be:
            raise be

        return local_path

    def create(self, deployment_instance):
        try:
            self._serving_api.put(deployment_instance)
            print("Deployment created, explore it at " + deployment_instance.get_url())
        except RestAPIError as re:
            raise_err = True
            if re.error_code == ModelServingException.ERROR_CODE_DUPLICATED_ENTRY:
                msg = "Deployment with the same name already exists"
                existing_deployment = self._serving_api.get(deployment_instance.name)
                if (
                    existing_deployment.model_name == deployment_instance.model_name
                    and existing_deployment.model_version
                    == deployment_instance.model_version
                ):  # if same name and model version, retrieve existing deployment
                    print(msg + ". Getting existing deployment...")
                    print("To create a new deployment choose a different name.")
                    deployment_instance.update_from_response_json(
                        existing_deployment.to_dict()
                    )
                    raise_err = False
                else:  # otherwise, raise an exception
                    print(", but it is serving a different model version.")
                    print("Please, choose a different name.")

            if raise_err:
                raise re

        if deployment_instance.is_stopped():
            print("Before making predictions, start the deployment by using `.start()`")

    def update(self, deployment_instance, await_update):
        state = deployment_instance.get_state()
        if state is None:
            return

        if state.status == PREDICTOR_STATE.STATUS_STARTING:
            # if starting, it cannot be updated yet
            raise ModelServingException(
                "Deployment is starting, please wait until it is running before applying changes. \n"
                + "Check the current status by using `.get_state()` or explore the server logs using `.get_logs()`"
            )
        if (
            state.status == PREDICTOR_STATE.STATUS_RUNNING
            or state.status == PREDICTOR_STATE.STATUS_IDLE
            or state.status == PREDICTOR_STATE.STATUS_FAILED
        ):
            # if running, it's fine
            self._serving_api.put(deployment_instance)
            print("Deployment updated, applying changes to running instances...")
            state = self._poll_deployment_status(  # wait for status
                deployment_instance, PREDICTOR_STATE.STATUS_RUNNING, await_update
            )
            if state is not None:
                if state.status == PREDICTOR_STATE.STATUS_RUNNING:
                    print("Running instances updated successfully")
            return
        if state.status == PREDICTOR_STATE.STATUS_UPDATING:
            # if updating, it cannot be updated yet
            raise ModelServingException(
                "Deployment is updating, please wait until it is running before applying changes. \n"
                + "Check the current status by using `.get_state()` or explore the server logs using `.get_logs()`"
            )
        if state.status == PREDICTOR_STATE.STATUS_STOPPING:
            # if stopping, it cannot be updated yet
            raise ModelServingException(
                "Deployment is stopping, please wait until it is stopped before applying changes"
            )
        if (
            state.status == PREDICTOR_STATE.STATUS_CREATING
            or state.status == PREDICTOR_STATE.STATUS_CREATED
            or state.status == PREDICTOR_STATE.STATUS_STOPPED
        ):
            # if stopped, it's fine
            self._serving_api.put(deployment_instance)
            print("Deployment updated, explore it at " + deployment_instance.get_url())
            return

        raise ValueError("Unknown deployment status: " + state.status)

    def save(self, deployment_instance, await_update: int):
        if deployment_instance.id is None:
            # if new deployment
            self.create(deployment_instance)
            return

        # if existing deployment
        self.update(deployment_instance, await_update)

    def delete(self, deployment_instance, force=False):
        state = deployment_instance.get_state()
        if state is None:
            return

        if (
            not force
            and state.status != PREDICTOR_STATE.STATUS_STOPPED
            and state.status != PREDICTOR_STATE.STATUS_CREATED
        ):
            raise ModelServingException(
                "Deployment not stopped, please stop it first by using `.stop()` or check its status with .get_state()"
            )

        self._serving_api.delete(deployment_instance)
        print("Deployment deleted successfully")

    def get_state(self, deployment_instance):
        try:
            state = self._serving_api.get_state(deployment_instance)
        except RestAPIError as re:
            if re.error_code == ModelServingException.ERROR_CODE_SERVING_NOT_FOUND:
                raise ModelServingException("Deployment not found") from re
            raise re
        deployment_instance._predictor._set_state(state)
        return state

    def get_logs(self, deployment_instance, component, tail):
        state = self.get_state(deployment_instance)
        if state is None:
            return

        if state.status == PREDICTOR_STATE.STATUS_STOPPING:
            print(
                "Deployment is stopping, explore historical logs at "
                + deployment_instance.get_url()
            )
            return
        if state.status == PREDICTOR_STATE.STATUS_STOPPED:
            print(
                "Deployment not running, explore historical logs at "
                + deployment_instance.get_url()
            )
            return
        if state.status == PREDICTOR_STATE.STATUS_STARTING:
            print("Deployment is starting, server logs might not be ready yet")

        print(
            "Explore all the logs and filters in the Kibana logs at "
            + deployment_instance.get_url(),
            end="\n\n",
        )

        return self._serving_api.get_logs(deployment_instance, component, tail)

    # Model inference

    def predict(
        self,
        deployment_instance,
        data: Union[Dict, List[InferInput]],
        inputs: Union[Dict, List[Dict]],
    ):
        # validate user-provided payload
        if deployment_instance.model_server == PREDICTOR.MODEL_SERVER_VLLM:
            raise ModelServingException(
                "Inference requests to LLM deployments are not supported by the `predict` method. Please, use any OpenAI API-compatible client instead."
            )

        self._validate_inference_payload(deployment_instance.api_protocol, data, inputs)

        # build inference payload based on API protocol
        payload = self._build_inference_payload(
            deployment_instance.api_protocol, data, inputs
        )

        # if not KServe, send request through Hopsworks
        serving_tool = deployment_instance.predictor.serving_tool
        through_hopsworks = serving_tool != PREDICTOR.SERVING_TOOL_KSERVE
        try:
            return self._serving_api.send_inference_request(
                deployment_instance, payload, through_hopsworks
            )
        except RestAPIError as re:
            if (
                re.response.status_code == RestAPIError.STATUS_CODE_NOT_FOUND
                or re.error_code
                == ModelServingException.ERROR_CODE_DEPLOYMENT_NOT_RUNNING
            ):
                raise ModelServingException(
                    "Deployment not created or running. If it is already created, start it by using `.start()` or check its status with .get_state()"
                ) from re

            re.args = (
                re.args[0] + "\n\n Check the model server logs by using `.get_logs()`",
            )
            raise re

    def _validate_inference_payload(
        self,
        api_protocol,
        data: Union[Dict, List[InferInput]],
        inputs: Union[Dict, List[Dict]],
    ):
        """Validates the user-provided inference payload. Either data or inputs parameter is expected, but both cannot be provided together."""
        # check null inputs
        if data is not None and inputs is not None:
            raise ModelServingException(
                "Inference data and inputs parameters cannot be provided together."
            )
        # check data or inputs
        if data is not None:
            self._validate_inference_data(api_protocol, data)
        else:
            self._validate_inference_inputs(api_protocol, inputs)

    def _validate_inference_data(
        self, api_protocol, data: Union[Dict, List[InferInput]]
    ):
        """Validates the inference payload when provided through the `data` parameter. The data parameter contains the raw payload to be sent
        in the inference request and should have the corresponding type and format depending on the API protocol.
        For the REST protocol, data should be a dictionary. For GRPC protocol, one or more InferInput objects is expected.
        """
        if api_protocol == IE.API_PROTOCOL_REST:  # REST protocol
            if isinstance(data, Dict):
                if "instances" not in data and "inputs" not in data:
                    raise ModelServingException(
                        "Inference data is missing 'instances' key."
                    )

                payload = data["instances"] if "instances" in data else data["inputs"]
                if not isinstance(payload, List):
                    raise ModelServingException(
                        "Instances field should contain a 2-dim list."
                    )
                elif len(payload) == 0:
                    raise ModelServingException(
                        "Inference data cannot contain an empty list."
                    )
                elif not isinstance(payload[0], List):
                    raise ModelServingException(
                        "Instances field should contain a 2-dim list."
                    )
                elif len(payload[0]) == 0:
                    raise ModelServingException(
                        "Inference data cannot contain an empty list."
                    )
            else:  # not Dict
                if isinstance(data, InferInput) or (
                    isinstance(data, List) and isinstance(data[0], InferInput)
                ):
                    raise ModelServingException(
                        "Inference data cannot contain `InferInput` for deployments with gRPC protocol disabled. Use a dictionary instead."
                    )
                raise ModelServingException(
                    "Inference data must be a dictionary. Otherwise, use the `inputs` parameter."
                )

        else:  # gRPC protocol
            if isinstance(data, Dict):
                raise ModelServingException(
                    "Inference data cannot be a dictionary for deployments with gRPC protocol enabled. "
                    "Create a `InferInput` object or use the `inputs` parameter instead."
                )
            elif isinstance(data, List):
                if len(data) == 0:
                    raise ModelServingException(
                        "Inference data cannot contain an empty list."
                    )
                if not isinstance(data[0], InferInput):
                    raise ModelServingException(
                        "Inference data must contain a list of `InferInput` objects. Otherwise, use the `inputs` parameter."
                    )
            else:
                raise ModelServingException(
                    "Inference data must contain a list of `InferInput` objects for deployments with gRPC protocol enabled."
                )

    def _validate_inference_inputs(
        self, api_protocol, inputs: Union[Dict, List[Dict]], recursive_call=False
    ):
        """Validates the inference payload when provided through the `inputs` parameter. The inputs parameter contains only the payload values,
        which will be parsed when building the request payload. It can be either a dictionary or a list.
        """
        if isinstance(inputs, List):
            if len(inputs) == 0:
                raise ModelServingException("Inference inputs cannot be an empty list.")
            else:
                self._validate_inference_inputs(
                    api_protocol, inputs[0], recursive_call=True
                )
        elif isinstance(inputs, InferInput):
            raise ModelServingException(
                "Inference inputs cannot be of type `InferInput`. Use the `data` parameter instead."
            )
        elif isinstance(inputs, Dict):
            required_keys = ("name", "shape", "datatype", "data")
            if api_protocol == IE.API_PROTOCOL_GRPC and not all(
                k in inputs for k in required_keys
            ):
                raise ModelServingException(
                    f"Inference inputs is missing one or more keys. Required keys are [{', '.join(required_keys)}]."
                )
        elif not recursive_call or (api_protocol == IE.API_PROTOCOL_GRPC):
            # if it is the first call to this method, inputs have an invalid type/format
            # if GRPC protocol is used, only Dict type is valid for the input values
            raise ModelServingException(
                "Inference inputs type is not valid. Supported types are dictionary and list."
            )

    def _build_inference_payload(
        self,
        api_protocol,
        data: Union[Dict, List[InferInput]],
        inputs: Union[Dict, List[Dict]],
    ):
        """Build the inference payload for an inference request. If the 'data' parameter is provided, this method ensures
        it has the correct format depending on the API protocol. Otherwise, if the 'inputs' parameter is provided, this method
        builds the correct request payload depending on the API protocol.
        """
        if data is not None:
            # data contains the raw payload (dict or InferInput), nothing needs to be changed
            return data
        else:  # parse inputs
            return self._parse_inference_inputs(api_protocol, inputs)

    def _parse_inference_inputs(
        self, api_protocol, inputs: Union[Dict, List[Dict]], recursive_call=False
    ):
        if api_protocol == IE.API_PROTOCOL_REST:  # REST protocol
            if not isinstance(inputs, List):
                data = {"instances": [[inputs]]}  # wrap inputs in a 2-dim list
            else:
                data = {"instances": inputs}  # use given inputs list by default
                # check depth of the list: at least two levels are required for batch inference
                # if the content is neither a list or dict, wrap it in an additional list
                for i in inputs:
                    if not isinstance(i, List) and not isinstance(i, Dict):
                        # if there are no two levels, wrap inputs in a list
                        data = {"instances": [inputs]}
                        break
        else:  # gRPC protocol
            if isinstance(inputs, Dict):  # Dict
                data = InferInput(
                    name=inputs["name"],
                    shape=inputs["shape"],
                    datatype=inputs["datatype"],
                    data=inputs["data"],
                    parameters=(
                        inputs["parameters"] if "parameters" in inputs else None
                    ),
                )
                if not recursive_call:
                    # if inputs is of type Dict, return a singleton
                    data = [data]

            else:  # List[Dict]
                data = inputs
                for index, inputs_item in enumerate(inputs):
                    data[index] = self._parse_inference_inputs(
                        api_protocol, inputs_item, recursive_call=True
                    )

        return data



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import warnings

from hsfs import engine, training_dataset_feature
from hsfs.constructor import query
from hsfs.core import (
    tags_api,
    training_dataset_api,
)


class TrainingDatasetEngine:
    OVERWRITE = "overwrite"
    APPEND = "append"
    ENTITY_TYPE = "trainingdatasets"

    def __init__(self, feature_store_id):
        self._feature_store_id = feature_store_id

        self._training_dataset_api = training_dataset_api.TrainingDatasetApi(
            feature_store_id
        )
        self._tags_api = tags_api.TagsApi(feature_store_id, self.ENTITY_TYPE)

    def save(self, training_dataset, features, user_write_options):
        if isinstance(features, query.Query):
            training_dataset._querydto = features
            training_dataset._features = [
                training_dataset_feature.TrainingDatasetFeature(
                    name=label_name, label=True
                )
                for label_name in training_dataset.label
            ]
        else:
            features = engine.get_instance().convert_to_default_dataframe(features)
            training_dataset._features = (
                engine.get_instance().parse_schema_training_dataset(features)
            )
            for label_name in training_dataset.label:
                for feature in training_dataset._features:
                    if feature.name == label_name:
                        feature.label = True

        if len(training_dataset.splits) > 0 and training_dataset.train_split is None:
            training_dataset.train_split = "train"
            warnings.warn(
                "Training dataset splits were defined but no `train_split` (the name of the split that is going to be "
                "used for training) was provided. Setting this property to `train`. ",
                stacklevel=1,
            )

        updated_instance = self._training_dataset_api.post(training_dataset)
        td_job = engine.get_instance().write_training_dataset(
            training_dataset, features, user_write_options, self.OVERWRITE
        )
        return updated_instance, td_job

    def insert(self, training_dataset, dataset, user_write_options, overwrite):
        return engine.get_instance().write_training_dataset(
            training_dataset,
            dataset,
            user_write_options,
            self.OVERWRITE if overwrite else self.APPEND,
        )

    def read(self, training_dataset, split, user_read_options):
        read_options = engine.get_instance().read_options(
            training_dataset.data_format, user_read_options
        )

        if split is not None:
            path = training_dataset.location + "/" + str(split)
        else:
            path = training_dataset.location + "/" + training_dataset.name

        return training_dataset.storage_connector.read(
            # always read from materialized dataset, not query object
            query=None,
            data_format=training_dataset.data_format,
            options=read_options,
            path=path,
        )

    def query(self, training_dataset, online, with_label, is_hive_query):
        fs_query = self._training_dataset_api.get_query(
            training_dataset, with_label, is_hive_query
        )

        if online:
            return fs_query.query_online

        # The offline queries could be referencing temporary tables
        # like external feature groups/hudi feature groups
        # Here we register those tables before returning the query to the user
        # In this way, if they execute the query, it will be valid
        fs_query.register_external()
        fs_query.register_hudi_tables(
            self._feature_store_id,
            None,  # No need to provide the feature store name for read operations
            {},
        )
        if fs_query.pit_query is not None:
            return fs_query.pit_query

        return fs_query.query

    def add_tag(self, training_dataset, name, value):
        """Attach a name/value tag to a training dataset."""
        self._tags_api.add(training_dataset, name, value)

    def delete_tag(self, training_dataset, name):
        """Remove a tag from a training dataset."""
        self._tags_api.delete(training_dataset, name)

    def get_tag(self, training_dataset, name):
        """Get tag with a certain name for a training dataset."""
        return self._tags_api.get(training_dataset, name)[name]

    def get_tags(self, training_dataset):
        """Get all tags for a training dataset."""
        return self._tags_api.get(training_dataset)

    def update_statistics_config(self, training_dataset):
        """Update the statistics configuration of a feature group."""
        self._training_dataset_api.update_metadata(
            training_dataset, training_dataset, "updateStatsConfig"
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
from typing import List, Optional, Union

from hopsworks_common import usage, util
from hopsworks_common.constants import ARTIFACT_VERSION, PREDICTOR_STATE
from hopsworks_common.constants import INFERENCE_ENDPOINTS as IE
from hsml.core import serving_api
from hsml.deployment import Deployment
from hsml.inference_batcher import InferenceBatcher
from hsml.inference_endpoint import InferenceEndpoint
from hsml.inference_logger import InferenceLogger
from hsml.model import Model
from hsml.predictor import Predictor
from hsml.resources import PredictorResources
from hsml.transformer import Transformer


class ModelServing:
    DEFAULT_VERSION = 1

    def __init__(
        self,
        project_name: str,
        project_id: int,
        **kwargs,
    ):
        self._project_name = project_name
        self._project_id = project_id

        self._serving_api = serving_api.ServingApi()

    @usage.method_logger
    def get_deployment_by_id(self, id: int) -> Optional[Deployment]:
        """Get a deployment by id from Model Serving.
        Getting a deployment from Model Serving means getting its metadata handle
        so you can subsequently operate on it (e.g., start or stop).

        !!! example
            ```python
            # login and get Hopsworks Model Serving handle using .login() and .get_model_serving()

            # get a deployment by id
            my_deployment = ms.get_deployment_by_id(1)
            ```

        # Arguments
            id: Id of the deployment to get.
        # Returns
            `Deployment`: The deployment metadata object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If unable to retrieve deployment from model serving.
        """

        return self._serving_api.get_by_id(id)

    @usage.method_logger
    def get_deployment(self, name: str = None) -> Optional[Deployment]:
        """Get a deployment by name from Model Serving.

        !!! example
            ```python
            # login and get Hopsworks Model Serving handle using .login() and .get_model_serving()

            # get a deployment by name
            my_deployment = ms.get_deployment('deployment_name')
            ```

        Getting a deployment from Model Serving means getting its metadata handle
        so you can subsequently operate on it (e.g., start or stop).

        # Arguments
            name: Name of the deployment to get.
        # Returns
            `Deployment`: The deployment metadata object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If unable to retrieve deployment from model serving.
        """

        if name is None and ("DEPLOYMENT_NAME" in os.environ):
            name = os.environ["DEPLOYMENT_NAME"]
        return self._serving_api.get(name)

    @usage.method_logger
    def get_deployments(
        self, model: Model = None, status: str = None
    ) -> List[Deployment]:
        """Get all deployments from model serving.
        !!! example
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            list_deployments = ms.get_deployment(my_model)

            for deployment in list_deployments:
                print(deployment.get_state())
            ```
        # Arguments
            model: Filter by model served in the deployments
            status: Filter by status of the deployments
        # Returns
            `List[Deployment]`: A list of deployments.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If unable to retrieve deployments from model serving.
        """

        model_name = model._get_default_serving_name() if model is not None else None
        if status is not None:
            self._validate_deployment_status(status)

        return self._serving_api.get_all(model_name, status)

    def _validate_deployment_status(self, status):
        statuses = list(util.get_members(PREDICTOR_STATE, prefix="STATUS"))
        status = status.upper()
        if status not in statuses:
            raise ValueError(
                "Deployment status '{}' is not valid. Possible values are '{}'".format(
                    status, ", ".join(statuses)
                )
            )
        return status

    def get_inference_endpoints(self) -> List[InferenceEndpoint]:
        """Get all inference endpoints available in the current project.

        # Returns
            `List[InferenceEndpoint]`: Inference endpoints for model inference
        """

        return self._serving_api.get_inference_endpoints()

    @usage.method_logger
    def create_predictor(
        self,
        model: Model,
        name: Optional[str] = None,
        artifact_version: Optional[str] = ARTIFACT_VERSION.CREATE,
        serving_tool: Optional[str] = None,
        script_file: Optional[str] = None,
        config_file: Optional[str] = None,
        resources: Optional[Union[PredictorResources, dict]] = None,
        inference_logger: Optional[Union[InferenceLogger, dict, str]] = None,
        inference_batcher: Optional[Union[InferenceBatcher, dict]] = None,
        transformer: Optional[Union[Transformer, dict]] = None,
        api_protocol: Optional[str] = IE.API_PROTOCOL_REST,
    ) -> Predictor:
        """Create a Predictor metadata object.

        !!! example
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            my_predictor = ms.create_predictor(my_model)

            my_deployment = my_predictor.deploy()
            ```

        !!! note "Lazy"
            This method is lazy and does not persist any metadata or deploy any model on its own.
            To create a deployment using this predictor, call the `deploy()` method.

        # Arguments
            model: Model to be deployed.
            name: Name of the predictor.
            artifact_version: Version number of the model artifact to deploy, `CREATE` to create a new model artifact
            or `MODEL-ONLY` to reuse the shared artifact containing only the model files.
            serving_tool: Serving tool used to deploy the model server.
            script_file: Path to a custom predictor script implementing the Predict class.
            config_file: Model server configuration file to be passed to the model deployment.
                It can be accessed via `CONFIG_FILE_PATH` environment variable from a predictor script.
                For LLM deployments without a predictor script, this file is used to configure the vLLM engine.
            resources: Resources to be allocated for the predictor.
            inference_logger: Inference logger configuration.
            inference_batcher: Inference batcher configuration.
            transformer: Transformer to be deployed together with the predictor.
            api_protocol: API protocol to be enabled in the deployment (i.e., 'REST' or 'GRPC'). Defaults to 'REST'.

        # Returns
            `Predictor`. The predictor metadata object.
        """

        if name is None:
            name = model._get_default_serving_name()

        return Predictor.for_model(
            model,
            name=name,
            artifact_version=artifact_version,
            serving_tool=serving_tool,
            script_file=script_file,
            config_file=config_file,
            resources=resources,
            inference_logger=inference_logger,
            inference_batcher=inference_batcher,
            transformer=transformer,
            api_protocol=api_protocol,
        )

    @usage.method_logger
    def create_transformer(
        self,
        script_file: Optional[str] = None,
        resources: Optional[Union[PredictorResources, dict]] = None,
    ) -> Transformer:
        """Create a Transformer metadata object.

        !!! example
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Dataset API instance
            dataset_api = project.get_dataset_api()

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            # create my_transformer.py Python script
            class Transformer(object):
                def __init__(self):
                    ''' Initialization code goes here '''
                    pass

                def preprocess(self, inputs):
                    ''' Transform the requests inputs here. The object returned by this method will be used as model input to make predictions. '''
                    return inputs

                def postprocess(self, outputs):
                    ''' Transform the predictions computed by the model before returning a response '''
                    return outputs

            uploaded_file_path = dataset_api.upload("my_transformer.py", "Resources", overwrite=True)
            transformer_script_path = os.path.join("/Projects", project.name, uploaded_file_path)

            my_transformer = ms.create_transformer(script_file=uploaded_file_path)

            # or

            from hsml.transformer import Transformer

            my_transformer = Transformer(script_file)
            ```

        !!! example "Create a deployment with the transformer"
            ```python

            my_predictor = ms.create_predictor(transformer=my_transformer)
            my_deployment = my_predictor.deploy()

            # or
            my_deployment = ms.create_deployment(my_predictor, transformer=my_transformer)
            my_deployment.save()
            ```

        !!! note "Lazy"
            This method is lazy and does not persist any metadata or deploy any transformer. To create a deployment using this transformer, set it in the `predictor.transformer` property.

        # Arguments
            script_file: Path to a custom predictor script implementing the Transformer class.
            resources: Resources to be allocated for the transformer.

        # Returns
            `Transformer`. The transformer metadata object.
        """

        return Transformer(script_file=script_file, resources=resources)

    @usage.method_logger
    def create_deployment(
        self,
        predictor: Predictor,
        name: Optional[str] = None,
        environment: Optional[str] = None,
    ) -> Deployment:
        """Create a Deployment metadata object.

        !!! example
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            my_predictor = ms.create_predictor(my_model)

            my_deployment = ms.create_deployment(my_predictor)
            my_deployment.save()
            ```

        !!! example "Using the model object"
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            my_deployment = my_model.deploy()

            my_deployment.get_state().describe()
            ```

        !!! example "Using the Model Serving handle"
            ```python
            # login into Hopsworks using hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            # get Hopsworks Model Serving handle
            ms = project.get_model_serving()

            my_predictor = ms.create_predictor(my_model)

            my_deployment = my_predictor.deploy()

            my_deployment.get_state().describe()
            ```

        !!! note "Lazy"
            This method is lazy and does not persist any metadata or deploy any model. To create a deployment, call the `save()` method.

        # Arguments
            predictor: predictor to be used in the deployment
            name: name of the deployment
            environment: The inference environment to use

        # Returns
            `Deployment`. The deployment metadata object.
        """

        return Deployment(predictor=predictor, name=name, environment=environment)

    @property
    def project_name(self):
        """Name of the project in which Model Serving is located."""
        return self._project_name

    @property
    def project_path(self):
        """Path of the project the registry is connected to."""
        return "/Projects/{}".format(self._project_name)

    @property
    def project_id(self):
        """Id of the project in which Model Serving is located."""
        return self._project_id

    def __repr__(self):
        return f"ModelServing(project: {self._project_name!r})"



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import feature, feature_group


class TestFeature:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["feature"]["get"]["response"]

        # Act
        f = feature.Feature.from_response_json(json)

        # Assert
        assert f.name == "intt"
        assert f.type == "int"
        assert f.description == "test_description"
        assert f.primary is True
        assert f.partition is False
        assert f.hudi_precombine_key is True
        assert f.online_type == "int"
        assert f.default_value == "1"  # default value should be specified as string
        assert f._feature_group_id == 15
        assert not f.on_demand

    def test_from_response_json_on_demand(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["feature"]["get_on_demand"]["response"]

        # Act
        f = feature.Feature.from_response_json(json)

        # Assert
        assert f.name == "intt"
        assert f.type == "int"
        assert f.description == "test_description"
        assert f.primary is True
        assert f.partition is False
        assert f.hudi_precombine_key is True
        assert f.online_type == "int"
        assert f.default_value == "1"  # default value should be specified as string
        assert f._feature_group_id == 15
        assert f.on_demand

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["feature"]["get_basic_info"]["response"]

        # Act
        f = feature.Feature.from_response_json(json)

        # Assert
        assert f.name == "intt"
        assert f.type is None
        assert f.description is None
        assert f.primary is False
        assert f.partition is False
        assert f.hudi_precombine_key is False
        assert f.online_type is None
        assert f.default_value is None
        assert f._feature_group_id is None

    def test_like(self):
        # Arrange
        f = feature.Feature("feature_name")

        # Act
        filter_obj = f.like("max%")

        # Assert
        assert filter_obj._feature == f
        assert filter_obj._condition == filter_obj.LK
        assert filter_obj._value == "max%"

    def test_name_sanitizing(self):
        # Arrange
        spaced_name = "col 1"
        upper_name = "Col1"
        both = "Bravo Col"

        # Act
        spaced_feature = feature.Feature(name=spaced_name)
        upper_feature = feature.Feature(name=upper_name)
        both_feature = feature.Feature(name=both)

        # Assert
        assert spaced_feature.name == "col_1"
        assert upper_feature.name == "col1"
        assert both_feature.name == "bravo_col"

    def test_get_fully_qualified_feature_name_without_prefix_and_without_use_fqn(self):
        # Arrange
        f = feature.Feature("feature_name")
        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
            event_time="event_time",
            featurestore_name="test_fs",
        )

        # Act
        result = f._get_fully_qualified_feature_name(fg)

        # Assert
        assert result == "feature_name"

    def test_get_fully_qualified_feature_name_with_prefix_and_without_use_fqn(self):
        # Arrange
        f = feature.Feature("feature_name")
        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
            event_time="event_time",
            featurestore_name="test_fs",
        )

        # Act
        result = f._get_fully_qualified_feature_name(fg, prefix="prefix_")

        # Assert
        assert result == "prefix_feature_name"

    def test_get_fully_qualified_feature_name_without_prefix_and_with_use_fqn(self):
        # Arrange
        f = feature.Feature("feature_name", use_fully_qualified_name=True)
        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
            event_time="event_time",
            featurestore_name="test_fs",
        )

        # Act
        result = f._get_fully_qualified_feature_name(fg)

        # Assert
        assert result == "test_fs_test_1_feature_name"

    def test_get_fully_qualified_feature_name_with_prefix_and_with_use_fqn(self):
        # Arrange
        f = feature.Feature("feature_name", use_fully_qualified_name=True)
        fg = feature_group.FeatureGroup(
            name="test",
            version=1,
            featurestore_id=99,
            primary_key=[],
            partition_key=[],
            id=10,
            event_time="event_time",
            featurestore_name="test_fs",
        )

        # Act
        result = f._get_fully_qualified_feature_name(fg, prefix="prefix_")

        # Assert
        assert result == "test_fs_test_1_feature_name"
